# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 Carlo Pavan <carlopav@gmail.com>                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides support functions to edit Part objects."""
## @package gui_edit_part_objects
# \ingroup draftguitools
# \brief Provides support functions to edit Part objects.

__title__ = "FreeCAD Draft Edit Tool"
__author__ = (
    "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin, Carlo Pavan"
)
__url__ = "https://www.freecad.org"

## \addtogroup draftguitools
# @{
import FreeCAD as App
import DraftVecUtils

from draftguitools.gui_edit_base_object import GuiTools


class PartLineGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(obj.X1, obj.Y1, obj.Z1))
        editpoints.append(App.Vector(obj.X2, obj.Y2, obj.Z2))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.X1 = v.x
            obj.Y1 = v.y
            obj.Z1 = v.z
        elif node_idx == 1:
            obj.X2 = v.x
            obj.Y2 = v.y
            obj.Z2 = v.z


class PartBoxGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.Length, 0, 0))
        editpoints.append(App.Vector(0, obj.Width, 0))
        editpoints.append(App.Vector(0, 0, obj.Height))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.Base + v
        elif node_idx == 1:
            _vector = DraftVecUtils.project(v, App.Vector(1, 0, 0))
            obj.Length = _vector.Length
        elif node_idx == 2:
            _vector = DraftVecUtils.project(v, App.Vector(0, 1, 0))
            obj.Width = _vector.Length
        elif node_idx == 3:
            _vector = DraftVecUtils.project(v, App.Vector(0, 0, 1))
            obj.Height = _vector.Length


class PartCylinderGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.Radius, 0, 0))
        editpoints.append(App.Vector(0, 0, obj.Height))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.Base + v
        elif node_idx == 1:
            if v.Length > 0.0:
                obj.Radius = v.Length
        elif node_idx == 2:
            _vector = DraftVecUtils.project(v, App.Vector(0, 0, 1))
            obj.Height = _vector.Length


class PartConeGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.Radius1, 0, 0))
        editpoints.append(App.Vector(obj.Radius2, 0, obj.Height))
        editpoints.append(App.Vector(0, 0, obj.Height))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.Base + v
        elif node_idx == 1:
            obj.Radius1 = v.Length  # TODO: Perhaps better to project on the face?
        elif node_idx == 2:
            v.z = 0
            obj.Radius2 = v.Length  # TODO: Perhaps better to project on the face?
        elif node_idx == 3:  # Height is last to have the priority on the radius
            _vector = DraftVecUtils.project(v, App.Vector(0, 0, 1))
            obj.Height = _vector.Length


class PartSphereGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.Radius, 0, 0))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.Base + v
        elif node_idx == 1:
            if v.Length > 0.0:
                obj.Radius = v.Length  # TODO: Perhaps better to project on the face?


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import os
import Path
import Path.Op.Base as PathOp

from PySide.QtCore import QT_TRANSLATE_NOOP

__title__ = "CAM Custom Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "CAM Custom object and FreeCAD command"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


class ObjectCustom(PathOp.ObjectOp):
    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """customOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties

        enums = {
            "Source": [
                (translate("PathCustom", "Text"), "Text"),
                (translate("PathCustom", "File"), "File"),
            ],
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def opFeatures(self, obj):
        return PathOp.FeatureTool | PathOp.FeatureCoolant

    def initOperation(self, obj):
        obj.addProperty(
            "App::PropertyEnumeration",
            "Source",
            "Path",
            "Source of gcode (text, file, ...)",
        )

        obj.addProperty(
            "App::PropertyFile",
            "GcodeFile",
            "Path",
            "File containing gcode to be inserted",
        )

        obj.addProperty(
            "App::PropertyStringList",
            "Gcode",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The G-code to be inserted"),
        )

        # populate the property enumerations
        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

        obj.Proxy = self
        self.setEditorModes(obj)

    def onChanged(self, obj, prop):
        if prop == "Source":
            self.setEditorModes(obj)

        if prop == "Active" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def opOnDocumentRestored(self, obj):
        if not hasattr(obj, "Source"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "Source",
                "Path",
                "Source of gcode (text, file, ...)",
            )

        if not hasattr(obj, "GcodeFile"):
            obj.addProperty(
                "App::PropertyFile",
                "GcodeFile",
                "Path",
                "File containing gcode to be inserted",
            )

        # populate the property enumerations
        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

    def onDocumentRestore(self, obj):
        self.setEditorModes(self, obj)

    def setEditorModes(self, obj, features=None):
        if not hasattr(obj, "Source"):
            return

        if obj.Source == "Text":
            obj.setEditorMode("GcodeFile", 2)
            obj.setEditorMode("Gcode", 0)
        elif obj.Source == "File":
            obj.setEditorMode("GcodeFile", 0)
            obj.setEditorMode("Gcode", 2)

    def findGcodeFile(self, filename):
        if os.path.exists(filename):
            # probably absolute, just return
            return filename

        doc_path = os.path.dirname(FreeCAD.ActiveDocument.FileName)
        prospective_path = os.path.join(doc_path, filename)

        if os.path.exists(prospective_path):
            return prospective_path

    def opExecute(self, obj):
        self.commandlist.append(Path.Command("(Begin Custom)"))
        errorNumLines = []
        errorLines = []
        counter = 0

        if obj.Source == "Text" and obj.Gcode:
            for l in obj.Gcode:
                counter += 1
                try:
                    newcommand = Path.Command(str(l))
                    self.commandlist.append(newcommand)
                except ValueError:
                    errorNumLines.append(counter)
                    if len(errorLines) < 7:
                        errorLines.append(f"{counter}: {str(l).strip()}")
            if errorLines:
                Path.Log.warning(
                    translate("PathCustom", "Total invalid lines in Custom Text G-code: %s")
                    % len(errorNumLines)
                )

        elif obj.Source == "File" and len(obj.GcodeFile) > 0:
            gcode_file = self.findGcodeFile(obj.GcodeFile)

            # could not determine the path
            if not gcode_file:
                Path.Log.error(
                    translate("PathCustom", "Custom file %s could not be found.") % obj.GcodeFile
                )
            else:
                with open(gcode_file) as fd:
                    for l in fd.readlines():
                        counter += 1
                        try:
                            newcommand = Path.Command(str(l))
                            self.commandlist.append(newcommand)
                        except ValueError:
                            errorNumLines.append(counter)
                            if len(errorLines) < 7:
                                errorLines.append(f"{counter}: {str(l).strip()}")
                if errorLines:
                    Path.Log.warning(f'"{gcode_file}"')
                    Path.Log.warning(
                        translate("PathCustom", "Total invalid lines in Custom File G-code: %s")
                        % len(errorNumLines)
                    )

        if errorNumLines:
            Path.Log.warning(
                translate("PathCustom", "Please check lines: %s")
                % ", ".join(map(str, errorNumLines))
            )

            if len(errorLines) > 7:
                errorLines.append("...")
            Path.Log.warning("\n" + "\n".join(errorLines))

        self.commandlist.append(Path.Command("(End Custom)"))


def SetupProperties():
    setup = []
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Custom operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectCustom(obj, name, parentJob)
    return obj



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint electrostatic potential task panel for the document object"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr, André Kapelrud, Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package task_constraint_electrostaticpotential
#  \ingroup FEM
#  \brief task panel for constraint electrostatic potential object

from PySide import QtCore

import FreeCAD
import FreeCADGui
from femguiutils import selection_widgets

from femtools import membertools
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):

    def __init__(self, obj):
        super().__init__(obj)

        self.parameter_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/ElectrostaticPotential.ui"
        )

        # geometry selection widget
        # start with Solid in list!
        self._selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face", "Edge", "Vertex"], True, False
        )

        # form made from param and selection widget
        self.form = [self.parameter_widget, self._selectionWidget]

        analysis = obj.getParentGroup()
        self._mesh = None
        self._part = None
        if analysis is not None:
            self._mesh = membertools.get_single_member(analysis, "Fem::FemMeshObject")
        if self._mesh is not None:
            self._part = self._mesh.Shape
        self._partVisible = None
        self._meshVisible = None

        QtCore.QObject.connect(
            self.parameter_widget.ckb_electromagnetic,
            QtCore.SIGNAL("toggled(bool)"),
            self.electromagnetic_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.cb_boundary_condition,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.boundary_condition_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_potential,
            QtCore.SIGNAL("toggled(bool)"),
            self.potential_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_potential,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.potential_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_av,
            QtCore.SIGNAL("toggled(bool)"),
            self.av_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_re,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_re_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_av_1,
            QtCore.SIGNAL("toggled(bool)"),
            self.av_1_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_re_1,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_re_1_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_av_2,
            QtCore.SIGNAL("toggled(bool)"),
            self.av_2_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_re_2,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_re_2_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_av_3,
            QtCore.SIGNAL("toggled(bool)"),
            self.av_3_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_re_3,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_re_3_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_im,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_im_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_im_1,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_im_1_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_im_2,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_im_2_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_av_im_3,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.av_im_3_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_capacitance_body,
            QtCore.SIGNAL("toggled(bool)"),
            self.capacitance_body_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.spb_capacitance_body,
            QtCore.SIGNAL("valueChanged(int)"),
            self.capacitance_body_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_potential_constant,
            QtCore.SIGNAL("toggled(bool)"),
            self.potential_constant_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_electric_infinity,
            QtCore.SIGNAL("toggled(bool)"),
            self.electric_infinity_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_electric_flux_density,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.electric_flux_density_changed,
        )

        self.init_parameter_widget()

    def open(self):
        if self._mesh is not None and self._part is not None:
            self._meshVisible = self._mesh.ViewObject.isVisible()
            self._partVisible = self._part.ViewObject.isVisible()
            self._mesh.ViewObject.hide()
            self._part.ViewObject.show()

    def reject(self):
        self._restoreVisibility()
        self._selectionWidget.finish_selection()
        return super().reject()

    def accept(self):
        if self.obj.References != self._selectionWidget.references:
            self.obj.References = self._selectionWidget.references
        self._set_params()
        self._selectionWidget.finish_selection()
        self._restoreVisibility()
        return super().accept()

    def _restoreVisibility(self):
        if self._mesh is not None and self._part is not None:
            if self._meshVisible:
                self._mesh.ViewObject.show()
            else:
                self._mesh.ViewObject.hide()
            if self._partVisible:
                self._part.ViewObject.show()
            else:
                self._part.ViewObject.hide()

    def _get_params(self):
        self.potential = self.obj.Potential
        self.potential_enabled = self.obj.PotentialEnabled

        self.av_re = self.obj.AV_re
        self.av_re_1 = self.obj.AV_re_1
        self.av_re_2 = self.obj.AV_re_2
        self.av_re_3 = self.obj.AV_re_3
        self.av_im = self.obj.AV_im
        self.av_im_1 = self.obj.AV_im_1
        self.av_im_2 = self.obj.AV_im_2
        self.av_im_3 = self.obj.AV_im_3

        self.av_enabled = self.obj.EnableAV
        self.av_1_enabled = self.obj.EnableAV_1
        self.av_2_enabled = self.obj.EnableAV_2
        self.av_3_enabled = self.obj.EnableAV_3

        self.boundary_condition = self.obj.BoundaryCondition
        self.potential_constant = self.obj.PotentialConstant
        self.electric_infinity = self.obj.ElectricInfinity
        self.capacitance_body_enabled = self.obj.CapacitanceBodyEnabled
        self.capacitance_body = self.obj.CapacitanceBody
        self.electric_flux_density = self.obj.ElectricFluxDensity

    def _set_params(self):
        self.obj.Potential = self.potential
        self.obj.PotentialEnabled = self.potential_enabled

        self.obj.AV_re = self.av_re
        self.obj.AV_re_1 = self.av_re_1
        self.obj.AV_re_2 = self.av_re_2
        self.obj.AV_re_3 = self.av_re_3
        self.obj.AV_im = self.av_im
        self.obj.AV_im_1 = self.av_im_1
        self.obj.AV_im_2 = self.av_im_2
        self.obj.AV_im_3 = self.av_im_3

        self.obj.EnableAV = self.av_enabled
        self.obj.EnableAV_1 = self.av_1_enabled
        self.obj.EnableAV_2 = self.av_2_enabled
        self.obj.EnableAV_3 = self.av_3_enabled

        self.obj.BoundaryCondition = self.boundary_condition
        self.obj.PotentialConstant = self.potential_constant
        self.obj.ElectricInfinity = self.electric_infinity
        self.obj.CapacitanceBodyEnabled = self.capacitance_body_enabled
        self.obj.CapacitanceBody = self.capacitance_body

        self.obj.ElectricFluxDensity = self.electric_flux_density

    def init_parameter_widget(self):
        self._get_params()

        self.parameter_widget.qsb_potential.setProperty("value", self.potential)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_potential).bind(
            self.obj, "Potential"
        )
        self.parameter_widget.ckb_potential.setChecked(self.potential_enabled)

        # scalar potential
        self.parameter_widget.qsb_av_re.setProperty("value", self.av_re)
        self.parameter_widget.qsb_av_re.setEnabled(self.av_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_re).bind(self.obj, "AV_re")
        self.parameter_widget.qsb_av_im.setProperty("value", self.av_im)
        self.parameter_widget.qsb_av_im.setEnabled(self.av_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_im).bind(self.obj, "AV_im")

        # vector potential
        self.parameter_widget.qsb_av_re_1.setProperty("value", self.av_re_1)
        self.parameter_widget.qsb_av_re_1.setEnabled(self.av_1_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_re_1).bind(self.obj, "AV_re_1")
        self.parameter_widget.qsb_av_re_2.setProperty("value", self.av_re_2)
        self.parameter_widget.qsb_av_re_2.setEnabled(self.av_2_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_re_2).bind(self.obj, "AV_re_2")
        self.parameter_widget.qsb_av_re_3.setProperty("value", self.av_re_3)
        self.parameter_widget.qsb_av_re_3.setEnabled(self.av_3_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_re_3).bind(self.obj, "AV_re_3")

        self.parameter_widget.qsb_av_im_1.setProperty("value", self.av_im_1)
        self.parameter_widget.qsb_av_im_1.setEnabled(self.av_1_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_im_1).bind(self.obj, "AV_im_1")
        self.parameter_widget.qsb_av_im_2.setProperty("value", self.av_im_2)
        self.parameter_widget.qsb_av_im_2.setEnabled(self.av_2_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_im_2).bind(self.obj, "AV_im_2")
        self.parameter_widget.qsb_av_im_3.setProperty("value", self.av_im_3)
        self.parameter_widget.qsb_av_im_3.setEnabled(self.av_3_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_av_im_3).bind(self.obj, "AV_im_3")

        self.parameter_widget.ckb_av.setChecked(self.av_enabled)
        self.parameter_widget.ckb_av_1.setChecked(self.av_1_enabled)
        self.parameter_widget.ckb_av_2.setChecked(self.av_2_enabled)
        self.parameter_widget.ckb_av_3.setChecked(self.av_3_enabled)

        self.parameter_widget.ckb_potential_constant.setChecked(self.potential_constant)

        self.parameter_widget.ckb_electric_infinity.setChecked(self.electric_infinity)

        self.parameter_widget.ckb_capacitance_body.setChecked(self.capacitance_body_enabled)
        self.parameter_widget.spb_capacitance_body.setProperty("value", self.capacitance_body)
        FreeCADGui.ExpressionBinding(self.parameter_widget.spb_capacitance_body).bind(
            self.obj, "CapacitanceBody"
        )

        self.parameter_widget.qsb_electric_flux_density.setProperty(
            "value", self.electric_flux_density
        )
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_electric_flux_density).bind(
            self.obj, "ElectricFluxDensity"
        )

        self.bc_enum = self.obj.getEnumerationsOfProperty("BoundaryCondition")
        index = self.bc_enum.index(self.boundary_condition)
        self.parameter_widget.cb_boundary_condition.addItems(self.bc_enum)
        self.parameter_widget.cb_boundary_condition.setCurrentIndex(index)

        # start with electromagnetic inputs hidden if no field is set
        if not (self.av_enabled or self.av_1_enabled or self.av_2_enabled or self.av_3_enabled):
            self.parameter_widget.ckb_electromagnetic.setChecked(False)

    def potential_changed(self, value):
        self.potential = value

    def potential_enabled_changed(self, value):
        self.potential_enabled = value
        self.parameter_widget.qsb_potential.setEnabled(value)

    def electromagnetic_enabled_changed(self, value):
        self.parameter_widget.gb_electromagnetic.setVisible(value)

    def av_enabled_changed(self, value):
        self.av_enabled = value
        self.parameter_widget.qsb_av_re.setEnabled(value)
        self.parameter_widget.qsb_av_im.setEnabled(value)

    def av_1_enabled_changed(self, value):
        self.av_1_enabled = value
        self.parameter_widget.qsb_av_re_1.setEnabled(value)
        self.parameter_widget.qsb_av_im_1.setEnabled(value)

    def av_2_enabled_changed(self, value):
        self.av_2_enabled = value
        self.parameter_widget.qsb_av_re_2.setEnabled(value)
        self.parameter_widget.qsb_av_im_2.setEnabled(value)

    def av_3_enabled_changed(self, value):
        self.av_3_enabled = value
        self.parameter_widget.qsb_av_re_3.setEnabled(value)
        self.parameter_widget.qsb_av_im_3.setEnabled(value)

    def av_re_changed(self, value):
        self.av_re = value

    def av_re_1_changed(self, value):
        self.av_re_1 = value

    def av_re_2_changed(self, value):
        self.av_re_2 = value

    def av_re_3_changed(self, value):
        self.av_re_3 = value

    def av_im_changed(self, value):
        self.av_im = value

    def av_im_1_changed(self, value):
        self.av_im_1 = value

    def av_im_2_changed(self, value):
        self.av_im_2 = value

    def av_im_3_changed(self, value):
        self.av_im_3 = value

    def potential_constant_changed(self, value):
        self.potential_constant = value

    def electric_infinity_changed(self, value):
        self.electric_infinity = value

    def capacitance_body_enabled_changed(self, value):
        self.capacitance_body_enabled = value
        self.parameter_widget.spb_capacitance_body.setEnabled(value)

    def capacitance_body_changed(self, value):
        self.capacitance_body = value
        self.parameter_widget.spb_capacitance_body.setValue(value)

    def electric_flux_density_changed(self, value):
        self.electric_flux_density = value

    def boundary_condition_changed(self, index):
        self.boundary_condition = self.bc_enum[index]
        if self.boundary_condition == "Dirichlet":
            self.parameter_widget.gb_neumann.setEnabled(False)
            self.parameter_widget.gb_dirichlet.setEnabled(True)
        elif self.boundary_condition == "Neumann":
            self.parameter_widget.gb_neumann.setEnabled(True)
            self.parameter_widget.gb_dirichlet.setEnabled(False)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import Path.Base.Generator.tapping as generator
import CAMTests.PathTestUtils as PathTestUtils

Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class TestPathTapGenerator(PathTestUtils.PathTestBase):
    def test00(self):
        """Test Basic Tap Generator Return"""
        v1 = FreeCAD.Vector(0, 0, 10)
        v2 = FreeCAD.Vector(0, 0, 0)

        e = Part.makeLine(v1, v2)

        result = generator.generate(e)

        self.assertTrue(type(result) is list)
        self.assertTrue(type(result[0]) is Path.Command)

        command = result[0]

        self.assertTrue(command.Name == "G84")
        self.assertTrue(command.Parameters["R"] == 10)
        self.assertTrue(command.Parameters["X"] == 0)
        self.assertTrue(command.Parameters["Y"] == 0)
        self.assertTrue(command.Parameters["Z"] == 0)

        # repeat must be > 0
        args = {"edge": e, "repeat": 0}
        self.assertRaises(ValueError, generator.generate, **args)

        # repeat must be integer
        args = {"edge": e, "repeat": 1.5}
        self.assertRaises(ValueError, generator.generate, **args)

    def test10(self):
        """Test edge alignment check"""
        v1 = FreeCAD.Vector(0, 10, 10)
        v2 = FreeCAD.Vector(0, 0, 0)
        e = Part.makeLine(v1, v2)
        self.assertRaises(ValueError, generator.generate, e)

        v1 = FreeCAD.Vector(0, 0, 0)
        v2 = FreeCAD.Vector(0, 0, 10)
        e = Part.makeLine(v1, v2)

        self.assertRaises(ValueError, generator.generate, e)

    def test30(self):
        """Test Basic Dwell Tap Generator Return"""
        v1 = FreeCAD.Vector(0, 0, 10)
        v2 = FreeCAD.Vector(0, 0, 0)

        e = Part.makeLine(v1, v2)

        result = generator.generate(e, dwelltime=0.5)

        self.assertTrue(type(result) is list)
        self.assertTrue(type(result[0]) is Path.Command)

        command = result[0]

        self.assertTrue(command.Name == "G84")
        self.assertTrue(command.Parameters["P"] == 0.5)

        # dwelltime should be a float
        args = {"edge": e, "dwelltime": 1}
        self.assertRaises(ValueError, generator.generate, **args)

    def test40(self):
        """Specifying retract height should set R parameter to specified value"""
        v1 = FreeCAD.Vector(0, 0, 10)
        v2 = FreeCAD.Vector(0, 0, 0)

        e = Part.makeLine(v1, v2)

        result = generator.generate(e, retractheight=20.0)

        command = result[0]

        self.assertTrue(command.Parameters["R"] == 20.0)

    def test41(self):
        """Not specifying retract height should set R parameter to Z position of start point"""
        v1 = FreeCAD.Vector(0, 0, 10)
        v2 = FreeCAD.Vector(0, 0, 0)

        e = Part.makeLine(v1, v2)

        result = generator.generate(e)

        command = result[0]

        self.assertTrue(command.Parameters["R"] == 10.0)

    def test44(self):
        """Non-float retract height should raise ValueError"""
        v1 = FreeCAD.Vector(0, 0, 10)
        v2 = FreeCAD.Vector(0, 0, 0)

        e = Part.makeLine(v1, v2)

        args = {"edge": e, "retractheight": 1}
        self.assertRaises(ValueError, generator.generate, **args)
        args = {"edge": e, "retractheight": "1"}
        self.assertRaises(ValueError, generator.generate, **args)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# *   This file has been modified from Sliptonic original Linux CNC post    *
# *   for use with a Dynapath 20 controller all changes and Modifications   *
# *   (c) Linden (Linden@aktfast.net) 2016                                  *
# *                                                                         *
# *   Additional changes 2020 adding arguments to control units, precision  *
# *   editor, header, etc. Now uses FreeCAD unit system to correctly        *
# *   set Feed rate - sliptonic                                             *
# *                                                                         *
# *                                                                         *
# ***************************************************************************


import FreeCAD
from FreeCAD import Units
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
import argparse
import datetime
import shlex
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a Tree Journyman 325 3 axis mill with Dynapath 20
controller in MM. This is a work in progress and very few of the functions
available on the Dynapath have been implemented at this time.
This postprocessor, once placed in the appropriate PathScripts folder,
can be used directly from inside FreeCAD, via the GUI importer or via python
scripts with:

Done
Coordinate Decimal limited to 3 places
Feed limited to hole number no decimal
Speed Limited to hole number no decimal
Machine Travel Limits Set to approximate values
Line numbers start at one and incremental by 1
Set preamble
Set postamble

To Do
Change G20 and G21 to G70 and G71 for metric or imperial units
Convert arcs to absolute
Strip comments and white spaces
Add file name in brackets limited to 8 alpha numeric no spaces all caps as
first line in file
Change Q to K For depth of peck on G83
Fix tool change
Limit comments length and characters to Uppercase, alpha numeric and
spaces add / prior to comments

import dynapath_post
dynapath_post.export(object,"/path/to/file.ncc","")
"""

parser = argparse.ArgumentParser(prog="dynapath_post", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--precision", default="3", help="number of digits of precision, default=3")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="G17\\nG90\\nG80\\nG40\\n"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M09\\nM05\\nG80\\nG40\\nG17\\nG90\\nM30\\n"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)


TOOLTIP_ARGS = parser.format_help()

now = datetime.datetime.now()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 1  # line number starting value

UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
MACHINE_NAME = "Tree MM"
CORNER_MIN = {"x": -340, "y": 0, "z": 0}
CORNER_MAX = {"x": 340, "y": -355, "z": -150}

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G17
G90
;G90.1 ;needed for simulation only
G80
G40
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M09
M05
G80
G40
G17
G90
M30
"""


# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global OUTPUT_DOUBLES

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return

    print("postprocessing...")
    gcode = ""

    # Find the machine.
    # The user my have overridden post processor defaults in the GUI.  Make sure we're using the current values in the Machine Def.
    myMachine = None
    for pathobj in objectslist:
        if hasattr(pathobj, "MachineName"):
            myMachine = pathobj.MachineName
        if hasattr(pathobj, "MachineUnits"):
            if pathobj.MachineUnits == "Metric":
                UNITS = "G21"
            else:
                UNITS = "G20"
    if myMachine is None:
        print("No machine found in this selection")

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: " + obj.Label + ")\n"
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: " + obj.Label + ")\n"
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

    # do the post_amble

    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    print("show editor: {}".format(SHOW_EDITOR))
    if FreeCAD.GuiUp and SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("done postprocessing.")

    gfile = pyopen(filename, "w")
    gfile.write(final)
    gfile.close()


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 1
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    global PRECISION
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    out = ""
    lastcommand = None

    precision_string = "." + str(PRECISION) + "f"

    # params = ['X','Y','Z','A','B','I','J','K','F','S'] #This list control the order of parameters
    params = ["X", "Y", "Z", "A", "B", "I", "J", "F", "S", "T", "Q", "R", "L"]

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        if OUTPUT_COMMENTS:
            out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        if not hasattr(pathobj, "Path"):  # groups might contain non-path things like stock.
            return out

        if OUTPUT_COMMENTS:
            out += linenumber() + "(Path: " + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:
            outstring = []
            command = c.Name
            outstring.append(command)
            # if modal: only print the command if it is not the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F":
                        speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                        if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                            outstring.append(
                                param
                                + format(
                                    float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                    precision_string,
                                )
                            )
                    elif param == "S":
                        outstring.append(param + format(c.Parameters[param], precision_string))
                    elif param == "T":
                        outstring.append(param + format(c.Parameters["T"], precision_string))
                    else:
                        pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                        outstring.append(
                            param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                        )

            # store the latest command
            lastcommand = command

            # Check for Tool Change:
            if command == "M6":
                if OUTPUT_COMMENTS:
                    out += linenumber() + "(begin toolchange)\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out = out.strip() + "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Electrostatics Elmer writer"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from FreeCAD import Units

from .. import sifio


class ESwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getElectrostaticSolver(self, equation):
        # check if we need to update the equation
        self._updateElectrostaticSolver(equation)
        # output the equation parameters
        s = self.write.createLinearSolver(equation)
        s["Equation"] = "Stat Elec Solver"  # equation.Name
        s["Procedure"] = sifio.FileAttr("StatElecSolve/StatElecSolver")
        s["Variable"] = self.write.getUniqueVarName("Potential")
        s["Variable DOFs"] = 1
        if equation.CalculateCapacitanceMatrix is True:
            s["Calculate Capacitance Matrix"] = equation.CalculateCapacitanceMatrix
            s["Capacitance Matrix Filename"] = equation.CapacitanceMatrixFilename
        if equation.CalculateElectricEnergy is True:
            s["Calculate Electric Energy"] = equation.CalculateElectricEnergy
        if equation.CalculateElectricField is True:
            s["Calculate Electric Field"] = equation.CalculateElectricField
        if equation.CalculateElectricFlux is True:
            s["Calculate Electric Flux"] = equation.CalculateElectricFlux
        if equation.CalculateSurfaceCharge is True:
            s["Calculate Surface Charge"] = equation.CalculateSurfaceCharge
        if equation.ConstantWeights is True:
            s["Constant Weights"] = equation.ConstantWeights
        s["Exec Solver"] = "Always"
        s["Optimize Bandwidth"] = True
        if equation.CalculateCapacitanceMatrix is False and (equation.PotentialDifference != 0.0):
            s["Potential Difference"] = equation.PotentialDifference
        s["Stabilize"] = equation.Stabilize
        return s

    def _updateElectrostaticSolver(self, equation):
        # updates older Electrostatic equations
        if not hasattr(equation, "CapacitanceMatrixFilename"):
            equation.addProperty(
                "App::PropertyFile",
                "CapacitanceMatrixFilename",
                "Electrostatic",
                (
                    "File where capacitance matrix is being saved\n"
                    "Only used if 'CalculateCapacitanceMatrix' is true"
                ),
                locked=True,
            )
            equation.CapacitanceMatrixFilename = "cmatrix.dat"
        if not hasattr(equation, "ConstantWeights"):
            equation.addProperty(
                "App::PropertyBool",
                "ConstantWeights",
                "Electrostatic",
                "Use constant weighting for results",
                locked=True,
            )
        if not hasattr(equation, "PotentialDifference"):
            equation.addProperty(
                "App::PropertyFloat",
                "PotentialDifference",
                "Electrostatic",
                (
                    "Potential difference in Volt for which capacitance is\n"
                    "calculated if 'CalculateCapacitanceMatrix' is false"
                ),
                locked=True,
            )
            equation.PotentialDifference = 0.0

    def handleElectrostaticConstants(self):
        permittivity = Units.Quantity(self.write.constsdef["PermittivityOfVacuum"]).getValueAs(
            "F/m"
        )
        self.write.constant("Permittivity Of Vacuum", permittivity)

    def handleElectrostaticMaterial(self, bodies):
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = obj.References[0][1] if obj.References else self.write.getAllBodies()
            for name in (n for n in refs if n in bodies):
                self.write.material(name, "Name", m["Name"])
                if "RelativePermittivity" in m:
                    self.write.material(
                        name, "Relative Permittivity", float(m["RelativePermittivity"])
                    )

    def handleElectrostaticBndConditions(self):
        for obj in self.write.getMember("Fem::ConstraintElectrostaticPotential"):
            if obj.References:
                for name in obj.References[0][1]:
                    # output the FreeCAD label as comment
                    if obj.Label:
                        self.write.boundary(name, "! FreeCAD Name", obj.Label)
                    if obj.BoundaryCondition == "Dirichlet":
                        if obj.PotentialEnabled:
                            potential = obj.Potential.getValueAs("V")
                            self.write.boundary(name, "Potential", potential)
                    elif obj.BoundaryCondition == "Neumann":
                        flux_density = obj.ElectricFluxDensity.getValueAs("C/m^2")
                        self.write.boundary(name, "Electric Flux", flux_density)
                    if obj.PotentialConstant:
                        self.write.boundary(name, "Potential Constant", True)
                    if obj.ElectricInfinity:
                        self.write.boundary(name, "Electric Infinity BC", True)
                    if obj.ElectricForcecalculation:
                        self.write.boundary(name, "Calculate Electric Force", True)
                    if obj.CapacitanceBodyEnabled:
                        self.write.boundary(name, "Capacitance Body", obj.CapacitanceBody)
                self.write.handled(obj)

        for obj in self.write.getMember("Fem::ConstraintElectricChargeDensity"):
            match obj.Mode:
                case "Interface":
                    density = obj.InterfaceChargeDensity
                case "Total Interface":
                    density = obj.Proxy.get_total_interface_density(obj)
                case _:
                    continue

            for feat, sub_elem in obj.References:
                for name in sub_elem:
                    self.write.boundary(name, "! FreeCAD Name", obj.Label)
                    self.write.boundary(
                        name,
                        "Surface Charge Density",
                        density.getValueAs("C/m^2"),
                    )
                    self.write.handled(obj)

    def handleElectrostaticBodyForces(self):
        for obj in self.write.getMember("Fem::ConstraintElectricChargeDensity"):
            match obj.Mode:
                case "Source":
                    density = obj.SourceChargeDensity
                case "Total Source":
                    density = obj.Proxy.get_total_source_density(obj)
                case _:
                    continue

            for feat, sub_elem in obj.References:
                for name in sub_elem:
                    self.write.bodyForce(name, "! FreeCAD Name", obj.Label)
                    self.write.bodyForce(name, "Charge Density", density.getValueAs("C/m^3"))
                    self.write.handled(obj)


##  @}



# /***************************************************************************
# *   Copyright (c) 2019 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

import FreeCAD


class TVObserver(object):
    def __init__(self):
        FreeCAD.addDocumentObserver(self)

    def stop(self):
        FreeCAD.removeDocumentObserver(self)

    def slotStartSaveDocument(self, doc, filepath):
        from . import TVStack

        TVStack._slotStartSaveDocument(doc)

    def slotFinishSaveDocument(self, doc, filepath):
        from . import TVStack

        TVStack._slotFinishSaveDocument(doc)

    def slotDeletedDocument(self, doc):
        from . import TVStack

        TVStack._slotDeletedDocument(doc)


# handle module reload
if "observer_singleton" in vars():
    observer_singleton.stop()

observer_singleton = TVObserver()



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""
#############################################################################
#
# ABOUT NETWORK MANAGER
#
# A wrapper around QNetworkAccessManager providing proxy-handling
# capabilities, and simplified access to submitting requests from any
# application thread.
#
#
# USAGE
#
# Once imported, this file provides access to a global object called
# AM_NETWORK_MANAGER. This is a QObject running on the main thread, but
# designed to be interacted with from any other application thread. It
# provides two principal methods: submit_unmonitored_get() and
# submit_monitored_get(). Use the unmonitored version for small amounts of
# data (suitable for caching in RAM, and without a need to show a progress
# bar during download), and the monitored version for larger amounts of data.
# Both functions take a URL, and return an integer index. That index allows
# tracking of the completed request by attaching to the signals completed(),
# progress_made(), and progress_complete(). All three provide, as the first
# argument to the signal, the index of the request the signal refers to.
# Code attached to those signals should filter them to look for the indices
# of the requests they care about. Requests may complete in any order.
#
# A secondary blocking interface is also provided, for very short network
# accesses: the blocking_get() function blocks until the network transmission
# is complete, directly returning a QByteArray object with the received data.
# Do not run on the main GUI thread!
"""

import threading
import os
import queue
import itertools
import tempfile
import time
from typing import Dict, List, Optional

import addonmanager_freecad_interface as fci
from addonmanager_preferences_migrations import migrate_proxy_settings_2025

from PySideWrapper import QtCore, QtNetwork, QtWidgets

translate = fci.translate


# This is the global instance of the NetworkManager that outside code
# should access
AM_NETWORK_MANAGER = None

# Added in Qt 5.15
if hasattr(QtNetwork.QNetworkRequest, "DefaultTransferTimeoutConstant"):
    timeoutConstant = QtNetwork.QNetworkRequest.DefaultTransferTimeoutConstant
    if hasattr(timeoutConstant, "value"):
        # Qt 6 changed the timeout constant to have a 'value' attribute.
        # The function setTransferTimeout does not accept
        # DefaultTransferTimeoutConstant of type
        # QtNetwork.QNetworkRequest.TransferTimeoutConstant any
        # longer but only an int.
        default_timeout = timeoutConstant.value
    else:
        # In Qt 5.15 we can use the timeoutConstant as is.
        default_timeout = timeoutConstant
else:
    default_timeout = 30000


class QueueItem:
    """A container for information about an item in the network queue."""

    def __init__(
        self,
        index: int,
        request: QtNetwork.QNetworkRequest,
        track_progress: bool,
        operation: QtNetwork.QNetworkAccessManager.Operation = QtNetwork.QNetworkAccessManager.GetOperation,
    ):
        self.index = index
        self.request = request
        self.original_url = request.url()
        self.track_progress = track_progress
        self.operation = operation


class NetworkManager(QtCore.QObject):
    """A single global instance of NetworkManager is instantiated and stored as
    AM_NETWORK_MANAGER. Outside threads should send GET requests to this class by
    calling the submit_unmonitored_request() or submit_monitored_request() function,
    as needed. See the documentation of those functions for details."""

    # Connect to complete for requests with no progress monitoring (e.g. small amounts of data)
    completed = QtCore.Signal(
        int, int, QtCore.QByteArray
    )  # Index, http response code, received data (if any)

    content_length = QtCore.Signal(int, int, int)

    # Connect to progress_made and progress_complete for large amounts of data, which get buffered into a temp file
    # That temp file should be deleted when your code is done with it
    progress_made = QtCore.Signal(int, int, int)  # Index, bytes read, total bytes (may be None)

    progress_complete = QtCore.Signal(int, int, os.PathLike)  # Index, http response code, filename

    __request_queued = QtCore.Signal()

    def __init__(self):
        super().__init__()

        self.counting_iterator = itertools.count()
        self.queue = queue.Queue()
        self.__last_started_index = 0
        self.__abort_when_found: List[int] = []
        self.replies: Dict[int, QtNetwork.QNetworkReply] = {}
        self.file_buffers = {}

        # We support an arbitrary number of threads using synchronous GET calls:
        self.synchronous_lock = threading.Lock()
        self.synchronous_complete: Dict[int, bool] = {}
        self.synchronous_result_data: Dict[int, QtCore.QByteArray] = {}

        # Only one size request at a time:
        self.download_size_lock = threading.Lock()

        # Make sure we exit nicely on quit
        if QtCore.QCoreApplication.instance() is not None:
            QtCore.QCoreApplication.instance().aboutToQuit.connect(self.__aboutToQuit)

        # Create the QNAM on this thread:
        self.QNAM = QtNetwork.QNetworkAccessManager()
        self.QNAM.proxyAuthenticationRequired.connect(self.__authenticate_proxy)
        self.QNAM.authenticationRequired.connect(self.__authenticate_resource)
        self.QNAM.setRedirectPolicy(QtNetwork.QNetworkRequest.ManualRedirectPolicy)

        qnam_cache = QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.CacheLocation)
        os.makedirs(qnam_cache, exist_ok=True)
        self.diskCache = QtNetwork.QNetworkDiskCache()
        self.diskCache.setCacheDirectory(qnam_cache)
        self.QNAM.setCache(self.diskCache)

        self.monitored_connections: List[int] = []
        self._setup_proxy()

        # A helper connection for our blocking interface
        self.completed.connect(self.__synchronous_process_completion)

        # Set up our worker connection
        self.__request_queued.connect(self.__setup_network_request)

    def _setup_proxy(self):
        """Set up the proxy based on user preferences or prompts on command line"""

        # Set up the proxy, if necessary:
        if fci.FreeCAD:
            migrate_proxy_settings_2025()
            proxy_type = fci.Preferences().get("proxy_type")
            proxy_host = fci.Preferences().get("proxy_host")
            proxy_port = fci.Preferences().get("proxy_port")
        else:
            proxy_type, proxy_host, proxy_port = self._setup_proxy_standalone()

        if proxy_type == "none":
            pass
        elif proxy_type == "system":
            query = QtNetwork.QNetworkProxyQuery(QtCore.QUrl("https://github.com/FreeCAD/FreeCAD"))
            proxy = QtNetwork.QNetworkProxyFactory.systemProxyForQuery(query)
            if proxy and proxy[0]:
                self.QNAM.setProxy(proxy[0])  # This may still be QNetworkProxy.NoProxy
        elif proxy_type == "custom":
            try:
                fci.Console.PrintMessage(f"Using proxy {proxy_host}:{proxy_port} \n")
                proxy = QtNetwork.QNetworkProxy(
                    QtNetwork.QNetworkProxy.HttpProxy, proxy_host, proxy_port
                )
                self.QNAM.setProxy(proxy)
            except Exception as e:
                fci.Console.PrintError(f"Error setting up proxy: {e}\n")

    def _setup_proxy_standalone(self):
        """If we are NOT running inside FreeCAD, prompt the user for proxy information"""
        print("Please select a proxy type:")
        print("1) No proxy")
        print("2) Use system proxy settings")
        print("3) Custom proxy settings")
        result = input("Choice: ")
        if result == "1":
            return "none", "", ""
        elif result == "2":
            return "system", "", ""
        elif result == "3":
            host = input("Enter proxy host: ")
            port = int(input("Enter proxy port: "))
        else:
            print("Defaulting to system proxy settings")
            return "system", "", ""
        return "custom", host, port

    def __aboutToQuit(self):
        """Called when the application is about to quit. Not currently used."""

    def __setup_network_request(self):
        """Get the next request off the queue and launch it."""
        try:
            item = self.queue.get_nowait()
            if item:
                if item.index in self.__abort_when_found:
                    self.__abort_when_found.remove(item.index)
                    return  # Do not do anything with this item, it's been aborted...
                if item.track_progress:
                    self.monitored_connections.append(item.index)
                self.__launch_request(item.index, item.request, item.operation)
        except queue.Empty:
            # Once the queue is empty, there's nothing left to do for now
            pass

    def __launch_request(
        self,
        index: int,
        request: QtNetwork.QNetworkRequest,
        operation: QtNetwork.QNetworkAccessManager.Operation,
    ) -> None:
        """Given a network request, ask the QNetworkAccessManager to begin processing it."""
        if operation == QtNetwork.QNetworkAccessManager.GetOperation:
            reply = self.QNAM.get(request)
        elif operation == QtNetwork.QNetworkAccessManager.HeadOperation:
            reply = self.QNAM.head(request)
        else:
            raise NotImplementedError(f"Unknown operation {operation.name}")
        self.replies[index] = reply

        self.__last_started_index = index
        reply.finished.connect(self.__reply_finished)
        reply.sslErrors.connect(self.__on_ssl_error)
        if index in self.monitored_connections:
            reply.readyRead.connect(self.__ready_to_read)
            reply.downloadProgress.connect(self.__download_progress)

    def query_download_size(self, url: str, timeout_ms: int = default_timeout):
        """A query to get the download size in the 'Content-Length' header, or zero if the server
        doesn't support it. Connect to the download_size signal to get the result when it arrives.
        """
        with self.download_size_lock:
            current_index = next(self.counting_iterator)  # A thread-safe counter
            item = QueueItem(
                current_index,
                self.__create_get_request(url, timeout_ms, disable_cache=True),
                track_progress=False,
                operation=QtNetwork.QNetworkAccessManager.HeadOperation,
            )
            self.queue.put(item)
            self.__request_queued.emit()
            return current_index

    def submit_unmonitored_get(
        self, url: str, timeout_ms: int = default_timeout, disable_cache: bool = False
    ) -> int:
        """Adds this request to the queue, and returns an index that can be used by calling code
        in conjunction with the completed() signal to handle the results of the call. All data is
        kept in memory, and the completed() call includes a direct handle to the bytes returned. It
        is not called until the data transfer has finished and the connection is closed."""

        current_index = next(self.counting_iterator)  # A thread-safe counter
        # Use a queue because we can only put things on the QNAM from the main event loop thread
        self.queue.put(
            QueueItem(
                current_index,
                self.__create_get_request(url, timeout_ms, disable_cache),
                track_progress=False,
            )
        )
        self.__request_queued.emit()
        return current_index

    def submit_monitored_get(
        self, url: str, timeout_ms: int = default_timeout, disable_cache: bool = False
    ) -> int:
        """Adds this request to the queue, and returns an index that can be used by calling code
        in conjunction with the progress_made() and progress_completed() signals to handle the
        results of the call. All data is cached to disk, and progress is reported periodically
        as the underlying QNetworkReply reports its progress. The progress_completed() signal
        contains a path to a temporary file with the stored data. Calling code should delete this
        file when done with it (or move it into its final place, etc.)."""

        current_index = next(self.counting_iterator)  # A thread-safe counter
        # Use a queue because we can only put things on the QNAM from the main event loop thread
        self.queue.put(
            QueueItem(
                current_index,
                self.__create_get_request(url, timeout_ms, disable_cache),
                track_progress=True,
            )
        )
        self.__request_queued.emit()
        return current_index

    def blocking_get_with_retries(
        self,
        url: str,
        timeout_ms: int = default_timeout,
        max_attempts: int = 3,
        delay_ms: int = 1000,
        disable_cache: bool = False,
    ):
        """Submits a GET request to the QNetworkAccessManager and blocks until it is complete. Do
        not use on the main GUI thread, it will prevent any event processing while it blocks.
        :url: The URL to fetch
        :timeout_ms: The timeout in milliseconds
        :max_attempts: The number of attempts to make if the request fails
        :delay_ms: The delay between attempts in milliseconds
        :returns: The response data, or None if the request failed after max_attempts attempts.
        """
        if max_attempts < 1:
            raise ValueError("max_attempts must be greater than 0")
        if timeout_ms < 1:
            raise ValueError("timeout_ms must be greater than 0")
        attempt = 0
        while True:
            attempt += 1
            p = self.blocking_get(url, timeout_ms)
            if p is not None or attempt >= max_attempts:
                return p
            if QtCore.QThread.currentThread().isInterruptionRequested():
                return None
            fci.Console.PrintWarning(
                f"Failed to get {url}, retrying in {delay_ms}ms... (attempt {attempt} of {max_attempts})\n"
            )
            time.sleep(delay_ms / 1000)

    def blocking_get(
        self, url: str, timeout_ms: int = default_timeout, disable_cache: bool = False
    ) -> Optional[QtCore.QByteArray]:
        """Submits a GET request to the QNetworkAccessManager and blocks until it is complete. Do
        not use on the main GUI thread, it will prevent any event processing while it blocks.
        :url: The URL to fetch
        :timeout_ms: The timeout in milliseconds
        :returns: The response data, or None if the request failed after max_attempts attempts.
        """

        current_index = next(self.counting_iterator)  # A thread-safe counter
        with self.synchronous_lock:
            self.synchronous_complete[current_index] = False

        self.queue.put(
            QueueItem(
                current_index,
                self.__create_get_request(url, timeout_ms, disable_cache),
                track_progress=False,
            )
        )
        self.__request_queued.emit()
        while True:
            if QtCore.QThread.currentThread().isInterruptionRequested():
                return None
            QtCore.QCoreApplication.processEvents()
            with self.synchronous_lock:
                if self.synchronous_complete[current_index]:
                    break

        with self.synchronous_lock:
            self.synchronous_complete.pop(current_index)
            if current_index in self.synchronous_result_data:
                return self.synchronous_result_data.pop(current_index)
            return None

    def __synchronous_process_completion(
        self, index: int, code: int, data: QtCore.QByteArray
    ) -> None:
        """Check the return status of a completed process, and handle its returned data (if
        any)."""
        with self.synchronous_lock:
            if index in self.synchronous_complete:
                if code == 200:
                    self.synchronous_result_data[index] = data
                else:
                    fci.Console.PrintWarning(
                        translate(
                            "AddonsInstaller",
                            "Addon Manager: Unexpected {} response from server",
                        ).format(code)
                        + "\n"
                    )
                self.synchronous_complete[index] = True

    @staticmethod
    def __create_get_request(
        url: str, timeout_ms: int, disable_cache: bool
    ) -> QtNetwork.QNetworkRequest:
        """Construct a network request to a given URL"""
        request = QtNetwork.QNetworkRequest(QtCore.QUrl(url))
        request.setAttribute(
            QtNetwork.QNetworkRequest.RedirectPolicyAttribute,
            QtNetwork.QNetworkRequest.ManualRedirectPolicy,
        )
        if disable_cache:
            request.setAttribute(QtNetwork.QNetworkRequest.CacheSaveControlAttribute, False)
            request.setAttribute(
                QtNetwork.QNetworkRequest.CacheLoadControlAttribute,
                QtNetwork.QNetworkRequest.AlwaysNetwork,
            )
        else:
            request.setAttribute(QtNetwork.QNetworkRequest.CacheSaveControlAttribute, True)
            request.setAttribute(
                QtNetwork.QNetworkRequest.CacheLoadControlAttribute,
                QtNetwork.QNetworkRequest.PreferNetwork,
            )
        if hasattr(request, "setTransferTimeout"):
            # Added in Qt 5.15
            # In Qt 5, the function setTransferTimeout seems to accept
            # DefaultTransferTimeoutConstant of type
            # PySide2.QtNetwork.QNetworkRequest.TransferTimeoutConstant,
            # whereas in Qt 6, the function seems to only accept an
            # integer.
            request.setTransferTimeout(timeout_ms)
        request.setRawHeader(b"User-Agent", b"FreeCAD AddonManager/1.0")
        return request

    def abort_all(self):
        """Abort ALL network calls in progress, including clearing the queue"""
        for reply in self.replies.values():
            if reply.isRunning():
                reply.abort()
        while True:
            try:
                self.queue.get()
                self.queue.task_done()
            except queue.Empty:
                break

    def abort(self, index: int):
        """Abort a specific request"""
        if index in self.replies and self.replies[index].isRunning():
            self.replies[index].abort()
        elif index < self.__last_started_index:
            # It's still in the queue. Mark it for later destruction.
            self.__abort_when_found.append(index)

    def __authenticate_proxy(
        self,
        reply: QtNetwork.QNetworkProxy,
        authenticator: QtNetwork.QAuthenticator,
    ):
        """If proxy authentication is required, attempt to authenticate. If the GUI is running this displays
        a window asking for credentials. If the GUI is not running, it prompts on the command line.
        """
        if fci.FreeCAD and fci.GuiUp:
            proxy_authentication = fci.loadUi(
                os.path.join(os.path.dirname(__file__), "proxy_authentication.ui")
            )
            proxy_authentication.setObjectName("AddonManager_ProxyAuthenticationDialog")
            # Show the right labels, etc.
            proxy_authentication.labelProxyAddress.setText(f"{reply.hostName()}:{reply.port()}")
            if authenticator.realm():
                proxy_authentication.labelProxyRealm.setText(authenticator.realm())
            else:
                proxy_authentication.labelProxyRealm.hide()
                proxy_authentication.labelRealmCaption.hide()
            result = proxy_authentication.exec()
            if result == QtWidgets.QDialogButtonBox.Ok:
                authenticator.setUser(proxy_authentication.lineEditUsername.text())
                authenticator.setPassword(proxy_authentication.lineEditPassword.text())
        else:
            username = input("Proxy username: ")
            import getpass

            password = getpass.getpass()
            authenticator.setUser(username)
            authenticator.setPassword(password)

    def __authenticate_resource(
        self,
        _reply: QtNetwork.QNetworkReply,
        _authenticator: QtNetwork.QAuthenticator,
    ):
        """Unused."""

    def __on_ssl_error(self, reply: str, errors: List[str] = None):
        """Called when an SSL error occurs: prints the error information."""
        fci.Console.PrintWarning(
            translate("AddonsInstaller", "Error with encrypted connection") + "\n:"
        )
        fci.Console.PrintWarning(reply)
        if errors is not None:
            for error in errors:
                fci.Console.PrintWarning(error)

    def __download_progress(self, bytesReceived: int, bytesTotal: int) -> None:
        """Monitors download progress and emits a progress_made signal"""
        sender = self.sender()
        if not sender:
            return
        for index, reply in self.replies.items():
            if reply == sender:
                self.progress_made.emit(index, bytesReceived, bytesTotal)
                return

    def __ready_to_read(self) -> None:
        """Called when data is available, this reads that data."""
        sender = self.sender()
        if not sender:
            return

        for index, reply in self.replies.items():
            if reply == sender:
                self.__data_incoming(index, reply)
                return

    def __data_incoming(self, index: int, reply: QtNetwork.QNetworkReply) -> None:
        """Read incoming data and attach it to a data object"""
        if not index in self.replies:
            # We already finished this reply, this is a vestigial signal
            return
        buffer = reply.readAll()
        if not index in self.file_buffers:
            f = tempfile.NamedTemporaryFile("wb", delete=False)
            self.file_buffers[index] = f
        else:
            f = self.file_buffers[index]
        try:
            f.write(buffer.data())
        except OSError as e:
            fci.Console.PrintError(f"Network Manager internal error: {str(e)}")

    def __reply_finished(self) -> None:
        """Called when a reply has been completed: this makes sure the data has been read and
        any notifications have been called."""
        reply: QtNetwork.QNetworkReply = self.sender()
        if not reply:
            # This can happen during a cancellation operation: silently do nothing
            return

        index = None
        for key, value in self.replies.items():
            if reply == value:
                index = key
                break
        if index is None:
            return

        response_code = reply.attribute(QtNetwork.QNetworkRequest.HttpStatusCodeAttribute)
        redirect_codes = [301, 302, 303, 305, 307, 308]
        if response_code in redirect_codes:  # This is a redirect
            timeout_ms = default_timeout
            if hasattr(reply, "request"):
                request = reply.request()
                if hasattr(request, "transferTimeout"):
                    timeout_ms = request.transferTimeout()
            new_url = reply.attribute(QtNetwork.QNetworkRequest.RedirectionTargetAttribute)
            disable_cache = not reply.request().attribute(
                QtNetwork.QNetworkRequest.CacheSaveControlAttribute
            )
            self.__launch_request(
                index,
                self.__create_get_request(new_url, timeout_ms, disable_cache=disable_cache),
                reply.operation(),
            )
            return  # The task is not done, so get out of this method now
        if reply.error() != QtNetwork.QNetworkReply.NetworkError.OperationCanceledError:
            # It this was not a timeout, make sure we mark the queue task done
            self.queue.task_done()
        if reply.error() == QtNetwork.QNetworkReply.NetworkError.NoError:
            if index in self.monitored_connections:
                # Make sure to read any remaining data
                self.__data_incoming(index, reply)
                self.monitored_connections.remove(index)
                f = self.file_buffers[index]
                f.close()
                self.progress_complete.emit(index, response_code, f.name)
            elif reply.operation() == QtNetwork.QNetworkAccessManager.HeadOperation:
                # The data we were trying to read was the ContentLengthHeader, so make sure that's what we emit
                data = reply.header(QtNetwork.QNetworkRequest.ContentLengthHeader)
                if data is None:
                    data = 0
                self.content_length.emit(index, response_code, data)
            else:
                data = reply.readAll()
                self.completed.emit(index, response_code, data)
        else:
            if index in self.monitored_connections:
                self.progress_complete.emit(index, response_code, "")
            elif reply.operation() == QtNetwork.QNetworkAccessManager.HeadOperation:
                self.content_length.emit(index, response_code, 0)
            else:
                self.completed.emit(index, response_code, None)
        self.replies.pop(index)


def InitializeNetworkManager():
    """Called once at the beginning of program execution to create the appropriate manager object"""
    global AM_NETWORK_MANAGER
    if AM_NETWORK_MANAGER is None:
        AM_NETWORK_MANAGER = NetworkManager()


def ForceReinitializeNetworkManager():
    """Called when the user changes the network settings, to force a re-initialization of the
    network manager."""
    global AM_NETWORK_MANAGER
    AM_NETWORK_MANAGER = NetworkManager()



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import math

import FreeCAD as App
import Part

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets


translate = App.Qt.translate

__title__ = "Assembly utilitary functions"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


def activePartOrAssembly():
    doc = Gui.ActiveDocument

    if doc is None or doc.ActiveView is None:
        return None
    activeAssembly = doc.ActiveView.getActiveObject("assembly")

    if activeAssembly:
        return activeAssembly

    return doc.ActiveView.getActiveObject("part")


def activeAssembly():
    active_assembly = activePartOrAssembly()
    if active_assembly is not None and active_assembly.isDerivedFrom("Assembly::AssemblyObject"):
        if active_assembly.ViewObject.isInEditMode():
            return active_assembly

    return None


def activePart():
    active_part = activePartOrAssembly()

    if active_part is not None and not active_part.isDerivedFrom("Assembly::AssemblyObject"):
        return active_part

    return None


def isAssemblyCommandActive():
    return activeAssembly() is not None and not Gui.Control.activeDialog()


def isDocTemporary(doc):
    # Guard against older versions of FreeCad which don't have the Temporary attribute
    try:
        temp = doc.Temporary
    except AttributeError:
        temp = False
    return temp


def assembly_has_at_least_n_parts(n):
    assembly = activeAssembly()
    if not assembly:
        assembly = activePart()
        if not assembly:
            return False
    i = number_of_components_in(assembly)
    return i >= n


def number_of_components_in(assembly):
    if not assembly:
        return 0
    i = 0
    for obj in assembly.Group:
        if isLinkGroup(obj):
            i = i + obj.ElementCount
            continue

        if obj.isDerivedFrom("Assembly::AssemblyObject") or obj.isDerivedFrom(
            "Assembly::AssemblyLink"
        ):
            i = i + number_of_components_in(obj)
            continue

        if obj.isDerivedFrom("App::Link"):
            obj = obj.getLinkedObject()

        if not obj.isDerivedFrom("App::GeoFeature"):
            continue

        if obj.isDerivedFrom("App::LocalCoordinateSystem"):
            continue

        i = i + 1

    return i


def isLink(obj):
    # If element count is not 0, then its a link group in which case the Link
    # is a container and it's the LinkElement that is linking to external doc.
    return (obj.TypeId == "App::Link" and obj.ElementCount == 0) or obj.TypeId == "App::LinkElement"


def isLinkGroup(obj):
    return obj.TypeId == "App::Link" and obj.ElementCount > 0


def getObject(ref):
    if len(ref) != 2 or ref[0] is None:
        return None
    subs = ref[1]
    if len(subs) < 1:
        return None
    sub_name = subs[0]

    # sub_name is "LinkOrAssembly1.LinkOrPart1.LinkOrBox.Edge16"
    # or           "LinkOrAssembly1.LinkOrPart1.LinkOrBody.pad.Edge16"
    # or           "Assembly.LinkOrAssembly1.LinkOrPart1.LinkOrBody.Local_CS.X"
    # We want either LinkOrBody or LinkOrBox or Local_CS.
    # Note since the ref now holds the moving part, sub_name can now be just the element name.
    # In this case the obj we need is the reference obj
    names = sub_name.split(".")
    names.insert(0, ref[0].Name)

    if len(names) < 2:
        return None

    doc = ref[0].Document
    for i, obj_name in enumerate(names):
        obj = doc.getObject(obj_name)

        if obj is None:
            return None

        # the last is the element name. So if we are at the last but one name, then it must be the selected
        if i == len(names) - 2:
            return obj

        if obj.TypeId in {"App::Part", "Assembly::AssemblyObject"} or isLinkGroup(obj):
            continue

        elif obj.isDerivedFrom("App::LocalCoordinateSystem"):
            # 2 cases possible, either we have the LCS itself: "part.LCS."
            # or we have a datum: "part.LCS.X_Axis"
            if i + 1 < len(names):
                obj2 = None
                for obji in obj.OutList:
                    if obji.Name == names[i + 1]:
                        obj2 = obji
                        break
                if obj2 and obj2.isDerivedFrom("App::DatumElement"):
                    return obj2

        elif obj.isDerivedFrom("App::DatumElement"):
            return obj

        elif obj.TypeId == "PartDesign::Body":
            return process_body(obj, obj, names, i)

        elif obj.isDerivedFrom("Part::Feature"):
            # primitive, fastener, gear ...
            return obj

        elif isLink(obj):
            linked_obj = obj.getLinkedObject()
            if linked_obj.TypeId == "PartDesign::Body":
                return process_body(linked_obj, obj, names, i)
            elif linked_obj.isDerivedFrom("Part::Feature"):
                return obj
            else:
                doc = linked_obj.Document
                continue

    return None


def process_body(body, returnObj, names, i):
    # If there's a next name, it could either be elementName, in which case we can return the Body.
    # But it could be a subobject like Sketch or datums.
    if i + 1 < len(names):
        obj2 = None
        for obji in body.OutList:
            if obji.Name == names[i + 1]:
                obj2 = obji
                break
        if obj2 and isBodySubObject(obj2):
            # obj2 can be a LCS or a Sketch. But it can also be a LCS's Axis or plane.
            if i + 2 < len(names):
                obj3 = None
                for obji in obj2.OutList:
                    if obji.Name == names[i + 2]:
                        obj3 = obji
                        break
                if obj3 and isBodySubObject(obj3):
                    return obj3
            return obj2
    return returnObj


def isBodySubObject(obj):
    return (
        obj.isDerivedFrom("Sketcher::SketchObject")
        or obj.isDerivedFrom("Part::Datum")
        or obj.isDerivedFrom("App::DatumElement")
        or obj.isDerivedFrom("App::LocalCoordinateSystem")
    )


def fixBodyExtraFeatureInSub(doc_name, sub_name):
    # If the sub_name that comes in has extra features in it, remove them.
    # For example :
    # "Part.Body.Pad.Edge2" -> "Part.Body.Edge2"
    # "Part.Body.Pad.Sketch." -> "Part.Body.Sketch."
    # "Body.Pad.Sketch." -> "Body.sketch."
    doc = App.getDocument(doc_name)
    names = sub_name.split(".")
    elt = names.pop()  # remove element

    bodyPassed = False
    new_sub_name = ""
    for obj_name in names:
        obj = doc.getObject(obj_name)
        if obj is None:
            return sub_name

        if bodyPassed and obj.isDerivedFrom("PartDesign::Feature"):
            continue  # we skip this name!

        if isLink(obj):
            obj = obj.getLinkedObject()
            doc = obj.Document

        if obj.TypeId == "PartDesign::Body":
            bodyPassed = True

        new_sub_name = new_sub_name + obj_name + "."

    new_sub_name = new_sub_name + elt  # Put back the element name

    return new_sub_name


def addTipNameToSub(ref):
    """
    Checks if the object referenced is a PartDesign::Body (or a Link to one).
    If so, it injects the Tip Name before the element name in the sub string.
    This is required for Gui.Selection to correctly highlight the geometry in the 3D View.
    Example: 'Body.Face1' -> 'Body.Pad.Face1'
    Example: 'Link.Face1' -> 'Link.Pad.Face1'
    """
    if not isRefValid(ref, 1):
        return ""

    sub = ref[1][0]

    # Resolve the actual object being pointed to
    obj = getObject(ref)

    target_body = None

    if obj:
        if obj.TypeId == "PartDesign::Body":
            target_body = obj
        elif isLink(obj):
            linked = obj.getLinkedObject()
            if linked and linked.TypeId == "PartDesign::Body":
                target_body = linked

    if target_body and target_body.Tip:
        split_sub = sub.split(".")
        if len(split_sub) > 0:
            element = split_sub.pop()
            if not element:  # Empty element name means the whole body is selected
                return sub

            split_sub.append(target_body.Tip.Name)
            split_sub.append(element)
            return ".".join(split_sub)

    return sub


# Deprecated. Kept for migrationScript.
def getObjectInPart(objName, part):
    if part is None:
        return None

    if part.Name == objName:
        return part

    if part.TypeId == "App::Link":
        part = part.getLinkedObject()

    if part.TypeId in {
        "App::Part",
        "Assembly::AssemblyObject",
        "App::DocumentObjectGroup",
        "PartDesign::Body",
    }:
        for obji in part.OutListRecursive:
            if obji.Name == objName:
                return obji

    return None


# Used by migrationScript.
def getRootPath(obj, part):
    sels = obj.Parents
    for sel in sels:
        rootObj = sel[0]
        # The part and the rootObj should be in the same doc
        if rootObj.Document.Name != part.Document.Name:
            continue

        path = sel[1]
        # we need to check that the part name is in the list.
        names = path.split(".")
        if part.Name not in names:
            continue

        # for bodies we need to add the tip to the path.
        if obj.TypeId == "PartDesign::Body":
            path = path + obj.Tip.Name + "."

        return rootObj, path

    return None, ""


# Return the jcs global placement
def getJcsGlobalPlc(jcsPlc, ref):
    obj_global_plc = getGlobalPlacement(ref)
    return obj_global_plc * jcsPlc


def getGlobalPlacement(ref, targetObj=None):
    if not isRefValid(ref, 1):
        return App.Placement()

    if targetObj is None:  # If no targetObj is given, we consider it's the getObject(ref)
        targetObj = getObject(ref)
        if targetObj is None:
            return App.Placement()

    rootObj = ref[0]
    subName = ref[1][0]
    # ref[0] is no longer the root object. Now it's the moving part.
    # So to get the correct global placement in case user transformed the assembly,
    # or if he has nested the assembly, we need to adjust it.
    # Example : Part / Assembly / Cylinder / Face1
    # ref is [(Cylinder, 'Face1')]
    parents = rootObj.Parents  # [(<Part object>, 'Assembly.Cylinder.')]
    if parents is not None and len(parents) == 1 and len(parents[0]) == 2:
        rootObj = parents[0][0]
        subName = parents[0][1] + subName

    return rootObj.getPlacementOf(subName, targetObj)


def isThereOneRootAssembly():
    for part in Gui.activeDocument().TreeRootObjects:
        if part.TypeId == "Assembly::AssemblyObject":
            return True
    return False


def getElementName(full_name):
    # full_name is "Assembly.Assembly1.Assembly2.Assembly3.Box.Edge16"
    # We want either Edge16.
    parts = full_name.split(".")

    if len(parts) < 1:
        # At minimum "Box.edge16". It shouldn't be shorter
        return ""

    # case of PartDesign datums : CoordinateSystem, point, line, plane
    if parts[-1] in {"X", "Y", "Z", "Point", "Line", "Plane"}:
        return ""

    return parts[-1]


def getObjsNamesAndElement(obj_name, sub_name):
    # if obj_name = "Assembly" and sub_name = "Assembly1.Assembly2.Assembly3.Box.Edge16"
    # this will return ["Assembly","Assembly1","Assembly2","Assembly3","Box"] and "Edge16"

    parts = sub_name.split(".")

    # The last part is always the element name even if empty
    element_name = parts[-1]

    # The remaining parts are object names
    obj_names = parts[:-1]
    obj_names.insert(0, obj_name)

    return obj_names, element_name


def getFullObjName(obj_name, sub_name):
    # if obj_name = "Assembly" and sub_name = "Assembly1.Assembly2.Assembly3.Box.Edge16"
    # this will return "Assembly.Assembly1.Assembly2.Assembly3.Box"
    objs_names, element_name = getObjsNamesAndElement(obj_name, sub_name)
    return ".".join(objs_names)


def getFullElementName(obj_name, sub_name):
    # if obj_name = "Assembly" and sub_name = "Assembly1.Assembly2.Assembly3.Box.Edge16"
    # this will return "Assembly.Assembly1.Assembly2.Assembly3.Box.Edge16"
    return obj_name + "." + sub_name


def extract_type_and_number(element_name):
    element_type = ""
    element_number = ""

    for char in element_name:
        if char.isalpha():
            # If the character is a letter, it's part of the type
            element_type += char
        elif char.isdigit():
            # If the character is a digit, it's part of the number
            element_number += char
        else:
            break

    if element_type and element_number:
        element_number = int(element_number)
        return element_type, element_number
    else:
        return None, None


def findElementClosestVertex(ref, mousePos):
    element_name = getElementName(ref[1][0])
    if element_name == "":
        return ""

    obj = getObject(ref)

    # We need mousePos to be in the same lcs as obj
    plc = App.Placement()
    plc.Base = mousePos
    global_plc = getGlobalPlacement(ref)
    plc = global_plc.inverse() * plc  # We make it relative to obj Origin
    plc = obj.Placement * plc  # Make plc in the same lcs as obj
    mousePos = plc.Base

    elt_type, elt_index = extract_type_and_number(element_name)

    if elt_type == "Vertex":
        return element_name

    elif elt_type == "Edge":
        edge = obj.Shape.getElement(element_name)
        curve = edge.Curve
        if curve.TypeId == "Part::GeomCircle":
            # For centers, as they are not shape vertexes, we return the element name.
            # For now we only allow selecting the center of arcs / circles.
            return element_name

        edge_points = getPointsFromVertexes(edge.Vertexes)

        if curve.TypeId == "Part::GeomLine":
            # For lines we allow users to select the middle of lines as well.
            line_middle = (edge_points[0] + edge_points[1]) * 0.5
            edge_points.append(line_middle)

        closest_vertex_index, _ = findClosestPointToMousePos(edge_points, mousePos)

        if curve.TypeId == "Part::GeomLine" and closest_vertex_index == 2:
            # If line center is closest then we have no vertex name to set so we put element name
            return element_name

        vertexes = edge.Vertexes

        vertex_name = findVertexNameInObject(vertexes[closest_vertex_index], obj)

        return vertex_name

    elif elt_type == "Face":
        face = obj.Shape.getElement(element_name)
        surface = face.Surface
        _type = surface.TypeId
        if _type == "Part::GeomSphere" or _type == "Part::GeomTorus":
            return element_name

        # Handle the circle/arc edges for their centers
        center_points = []
        center_points_edge_indexes = []
        edges = face.Edges

        for i, edge in enumerate(edges):
            curve = edge.Curve
            if curve.TypeId == "Part::GeomCircle" or curve.TypeId == "Part::GeomEllipse":
                center_points.append(curve.Location)
                center_points_edge_indexes.append(i)

            elif _type == "Part::GeomCylinder" and curve.TypeId == "Part::GeomBSplineCurve":
                # handle special case of 2 cylinder intersecting.
                for j, facej in enumerate(obj.Shape.Faces):
                    surfacej = facej.Surface
                    if (elt_index - 1) != j and surfacej.TypeId == "Part::GeomCylinder":
                        for edgej in facej.Edges:
                            if edgej.Curve.TypeId == "Part::GeomBSplineCurve":
                                if (
                                    edgej.CenterOfGravity == edge.CenterOfGravity
                                    and edgej.Length == edge.Length
                                ):
                                    center_points.append(edgej.CenterOfGravity)
                                    center_points_edge_indexes.append(i)

        if len(center_points) > 0:
            closest_center_index, closest_center_distance = findClosestPointToMousePos(
                center_points, mousePos
            )

        # Handle the face vertexes
        face_points = []

        if _type != "Part::GeomCylinder" and _type != "Part::GeomCone":
            face_points = getPointsFromVertexes(face.Vertexes)

        # We also allow users to select the center of gravity.
        if _type == "Part::GeomCylinder" or _type == "Part::GeomCone":
            centerOfG = face.CenterOfGravity - surface.Center
            centerPoint = surface.Center + centerOfG
            centerPoint = centerPoint + App.Vector().projectToLine(centerOfG, surface.Axis)
            face_points.append(centerPoint)
        else:
            face_points.append(face.CenterOfGravity)

        closest_vertex_index, closest_vertex_distance = findClosestPointToMousePos(
            face_points, mousePos
        )

        if len(center_points) > 0:
            if closest_center_distance < closest_vertex_distance:
                # Note the index here is the index within the face! Not the object.
                index = center_points_edge_indexes[closest_center_index] + 1
                return "Edge" + str(index)

        if _type == "Part::GeomCylinder" or _type == "Part::GeomCone":
            return element_name

        if closest_vertex_index == len(face.Vertexes):
            # If center of gravity then we have no vertex name to set so we put element name
            return element_name

        vertex_name = findVertexNameInObject(face.Vertexes[closest_vertex_index], obj)

        return vertex_name

    return ""


def getPointsFromVertexes(vertexes):
    points = []
    for vtx in vertexes:
        points.append(vtx.Point)
    return points


def findClosestPointToMousePos(candidates_points, mousePos):
    closest_point_index = None
    point_min_length = None

    for i, point in enumerate(candidates_points):
        length = (mousePos - point).Length
        if closest_point_index is None or length < point_min_length:
            closest_point_index = i
            point_min_length = length

    return closest_point_index, point_min_length


def findVertexNameInObject(vertex, obj):
    for i, _ in enumerate(obj.Shape.Vertexes):
        vertexName = "Vertex" + str(i + 1)

        if (
            obj.Shape.getElement(vertexName).Point == vertex.Point
        ):  # Don't use the vtx param from enumerate's 2nd output to avoid hasher warning
            return vertexName
    return ""


def color_from_unsigned(c):
    return [
        float(int((c >> 24) & 0xFF) / 255),
        float(int((c >> 16) & 0xFF) / 255),
        float(int((c >> 8) & 0xFF) / 255),
    ]


def getJointsOfType(asm, jointTypes):
    if not (
        asm.isDerivedFrom("Assembly::AssemblyObject") or asm.isDerivedFrom("Assembly::AssemblyLink")
    ):
        return []

    joints = []
    allJoints = asm.Joints
    for joint in allJoints:
        if joint.JointType in jointTypes:
            joints.append(joint)
    return joints


def getBomGroup(assembly):
    bom_group = None

    for obj in assembly.OutList:
        if obj.TypeId == "Assembly::BomGroup":
            bom_group = obj
            break

    if not bom_group:
        bom_group = assembly.newObject("Assembly::BomGroup", "Bills of Materials")

    return bom_group


def getJointGroup(assembly):
    joint_group = None

    for obj in assembly.OutList:
        if obj.TypeId == "Assembly::JointGroup":
            joint_group = obj
            break

    if not joint_group:
        joint_group = assembly.newObject("Assembly::JointGroup", "Joints")

    return joint_group


def getViewGroup(assembly):
    view_group = None

    for obj in assembly.OutList:
        if obj.TypeId == "Assembly::ViewGroup":
            view_group = obj
            break

    if not view_group:
        view_group = assembly.newObject("Assembly::ViewGroup", "Exploded Views")

    return view_group


def getSimulationGroup(assembly):
    sim_group = None

    for obj in assembly.OutList:
        if obj.TypeId == "Assembly::SimulationGroup":
            sim_group = obj
            break

    if not sim_group:
        sim_group = assembly.newObject("Assembly::SimulationGroup", "Simulations")

    return sim_group


def isAssemblyGrounded():
    assembly = activeAssembly()
    if not assembly:
        return False

    jointGroup = getJointGroup(assembly)

    for joint in jointGroup.Group:
        if hasattr(joint, "ObjectToGround"):
            return True

    return False


def removeObjAndChilds(obj):
    removeObjsAndChilds([obj])


def removeObjsAndChilds(objs):
    def addsubobjs(obj, toremoveset):
        if obj.TypeId == "App::Origin":  # Origins are already handled
            return

        toremoveset.add(obj)
        if obj.TypeId != "App::Link":
            for subobj in obj.OutList:
                addsubobjs(subobj, toremoveset)

    toremove = set()
    for obj in objs:
        addsubobjs(obj, toremove)

    for obj in toremove:
        if obj:
            obj.Document.removeObject(obj.Name)


# This function returns all the objects within the argument that can move:
# - Part::Features (outside of parts)
# - App::parts
# - App::Links to Part::Features or App::parts.
# It does not include Part::Features that are within App::Parts.
# It includes things inside Groups.
def getMovablePartsWithin(group, partsAsSolid=False):
    children = []
    if isLinkGroup(group):
        children = group.ElementList
    elif hasattr(group, "Group"):
        children = group.Group

    parts = []
    for obj in children:
        parts = parts + getSubMovingParts(obj, partsAsSolid)
    return parts


def getSubMovingParts(obj, partsAsSolid):
    if obj.isDerivedFrom("Part::Feature"):
        return [obj]

    elif obj.isDerivedFrom("App::Part"):
        objs = []
        if not partsAsSolid:
            objs = getMovablePartsWithin(obj)
        objs.append(obj)
        return objs

    elif isLinkGroup(obj) or obj.TypeId == "App::DocumentObjectGroup":
        return getMovablePartsWithin(obj)

    if isLink(obj):
        linked_obj = obj.getLinkedObject()
        if linked_obj.isDerivedFrom("App::Part") or linked_obj.isDerivedFrom("Part::Feature"):
            return [obj]

    return []


# Find the center of mass of a list of parts.
# Note it could be useful to move this to Measure mod.
def getCenterOfMass(parts):
    total_mass = 0
    total_com = App.Vector(0, 0, 0)

    for part in parts:
        mass, com = getObjMassAndCom(part)
        total_mass += mass
        total_com += com

    # After all parts are processed, calculate the overall center of mass
    if total_mass > 0:  # Avoid division by zero
        overall_com = total_com / total_mass
    else:
        overall_com = App.Vector(0, 0, 0)  # Default if no mass is found

    return overall_com


def getObjMassAndCom(obj, containingPart=None):
    link_global_plc = None

    if isLink(obj):
        link_global_plc = getGlobalPlacement(obj, containingPart)
        obj = obj.getLinkedObject()

    if obj.TypeId == "PartDesign::Body":
        obj = obj.Tip

    if obj.isDerivedFrom("Part::Feature"):
        mass = obj.Shape.Volume
        com = obj.Shape.CenterOfGravity
        # com takes into account obj placement, but not container placements.
        # So we make com relative to the obj :
        comPlc = App.Placement()
        comPlc.Base = com
        comPlc = obj.Placement.inverse() * comPlc
        # Then we make it relative to the origin of the doc
        global_plc = App.Placement()
        if link_global_plc is None:
            global_plc = getGlobalPlacement(obj, containingPart)
        else:
            global_plc = link_global_plc

        comPlc = global_plc * comPlc

        com = comPlc.Base * mass
        return mass, com

    elif (
        isLinkGroup(obj)
        or obj.isDerivedFrom("App::Part")
        or obj.isDerivedFrom("App::DocumentObjectGroup")
    ):
        if containingPart is None and obj.isDerivedFrom("App::Part"):
            containingPart = obj

        total_mass = 0
        total_com = App.Vector(0, 0, 0)

        if isLinkGroup(obj):
            children = obj.ElementList
        else:
            children = obj.Group

        for subObj in children:
            mass, com = getObjMassAndCom(subObj, containingPart)
            total_mass += mass
            total_com += com
        return total_mass, total_com
    return 0, App.Vector(0, 0, 0)


def getCenterOfBoundingBox(objs, refs):
    i = 0
    center = App.Vector()
    for obj, ref in zip(objs, refs):
        viewObject = obj.ViewObject
        if viewObject is None:
            continue
        boundingBox = viewObject.getBoundingBox()
        if boundingBox is None:
            continue
        bboxCenter = boundingBox.Center

        # bboxCenter does not take into account obj global placement
        plc = App.Placement(bboxCenter, App.Rotation())
        # change plc to be relative to the object placement.
        plc = obj.Placement.inverse() * plc
        # change plc to be relative to the origin of the document.
        global_plc = getGlobalPlacement(ref, obj)
        plc = global_plc * plc
        bboxCenter = plc.Base

        center = center + bboxCenter
        i = i + 1

    if i != 0:
        center = center / i
    return center


def findCylindersIntersection(obj, surface, edge, elt_index):
    for j, facej in enumerate(obj.Shape.Faces):
        surfacej = facej.Surface
        if (elt_index - 1) == j or surfacej.TypeId != "Part::GeomCylinder":
            continue

        for edgej in facej.Edges:
            if (
                edgej.Curve.TypeId == "Part::GeomBSplineCurve"
                and edgej.CenterOfGravity == edge.CenterOfGravity
                and edgej.Length == edge.Length
            ):
                # we need intersection between the 2 cylinder axis.
                line1 = Part.Line(surface.Center, surface.Center + surface.Axis)
                line2 = Part.Line(surfacej.Center, surfacej.Center + surfacej.Axis)

                res = line1.intersect(line2, Part.Precision.confusion())

                if res:
                    return App.Vector(res[0].X, res[0].Y, res[0].Z)
    return surface.Center


def openEditingPlacementDialog(obj, propName):
    task_placement = Gui.TaskPlacement()
    dialog = task_placement.form

    # Connect to the placement property
    task_placement.setPlacement(getattr(obj, propName))
    task_placement.setSelection([obj])
    task_placement.setPropertyName(propName)
    task_placement.bindObject()
    task_placement.setIgnoreTransactions(True)

    dialog.findChild(QtWidgets.QPushButton, "selectedVertex").hide()
    dialog.exec_()


def setPickableState(obj, state: bool):
    vobj = obj.ViewObject
    if hasattr(vobj, "Proxy"):
        proxy = vobj.Proxy
        if hasattr(proxy, "setPickableState"):
            proxy.setPickableState(state)


def setJointsPickableState(doc, state: bool):
    """Make all joints in document selectable (True) or unselectable (False) in 3D view"""
    for obj in doc.Objects:
        if obj.TypeId == "App::FeaturePython" and hasattr(obj, "JointType"):
            setPickableState(obj, state)


def applyOffsetToPlacement(plc, offset):
    plc.Base = plc.Base + plc.Rotation.multVec(offset)
    return plc


def applyRotationToPlacement(plc, angle):
    return applyRotationToPlacementAlongAxis(plc, angle, App.Vector(0, 0, 1))


def applyRotationToPlacementAlongAxis(plc, angle, axis):
    rot = plc.Rotation
    zRotation = App.Rotation(axis, angle)
    plc.Rotation = rot * zRotation
    return plc


def flipPlacement(plc):
    return applyRotationToPlacementAlongAxis(plc, 180, App.Vector(1, 0, 0))


def arePlacementSameDir(plc1, plc2):
    zAxis1 = plc1.Rotation.multVec(App.Vector(0, 0, 1))
    zAxis2 = plc2.Rotation.multVec(App.Vector(0, 0, 1))
    return zAxis1.dot(zAxis2) > 0


def arePlacementZParallel(plc1, plc2):
    zAxis1 = plc1.Rotation.multVec(App.Vector(0, 0, 1))
    zAxis2 = plc2.Rotation.multVec(App.Vector(0, 0, 1))
    return zAxis1.cross(zAxis2).Length < 1e-06


def removeTNPFromSubname(doc_name, obj_name, sub_name):
    rootObj = App.getDocument(doc_name).getObject(obj_name)
    resolved = rootObj.resolveSubElement(sub_name)
    element_name_TNP = resolved[1]
    element_name = resolved[2]

    # Preprocess the sub_name to remove the TNP string
    # We do this because after we need to add the vertex_name as well.
    # And the names will be resolved anyway after.
    if len(element_name_TNP.split(".")) == 2:
        names = sub_name.split(".")
        names.pop(-2)  # remove the TNP string
        sub_name = ".".join(names)

    return sub_name


"""
So here we want to find a placement that corresponds to a local coordinate system that would be placed at the selected vertex.
- obj is usually a App::Link to a PartDesign::Body, or primitive, fasteners. But can also be directly the object.1
- elt can be a face, an edge or a vertex.
- If elt is a vertex, then vtx = elt And placement is vtx coordinates without rotation.
- if elt is an edge, then vtx = edge start/end vertex depending on which is closer. If elt is an arc or circle, vtx can also be the center. The rotation is the plane normal to the line positioned at vtx. Or for arcs/circle, the plane of the arc.
- if elt is a plane face, vtx is the face vertex (to the list of vertex we need to add arc/circle centers) the closer to the mouse. The placement is the plane rotation positioned at vtx
- if elt is a cylindrical face, vtx can also be the center of the arcs of the cylindrical face.
"""


def findPlacement(ref, ignoreVertex=False):
    if not isRefValid(ref, 2):
        return App.Placement()
    obj = getObject(ref)
    if not obj:
        return App.Placement()

    elt = getElementName(ref[1][0])
    vtx = getElementName(ref[1][1])

    if not elt or not vtx:
        # case of whole parts such as PartDesign::Body or App/PartDesign::CordinateSystem/Point/Line/Plane.
        if obj.TypeId == "App::Line":
            if obj.Role == "X_Axis":
                return App.Placement(App.Vector(), App.Rotation(0.5, 0.5, 0.5, 0.5))
            if obj.Role == "Y_Axis":
                return App.Placement(App.Vector(), App.Rotation(0.5, 0.5, 0.5, 0.5))
            if obj.Role == "Z_Axis":
                return App.Placement(App.Vector(), App.Rotation(-0.5, 0.5, -0.5, 0.5))
        return App.Placement()

    plc = App.Placement()

    elt_type, elt_index = extract_type_and_number(elt)
    vtx_type, _ = extract_type_and_number(vtx)

    isLine = False

    if elt_type == "Vertex":
        vertex = get_element(obj.Shape, elt)
        if vertex is None:
            return App.Placement()
        plc.Base = (vertex.X, vertex.Y, vertex.Z)
    elif elt_type == "Edge":
        edge = get_element(obj.Shape, elt)
        if edge is None:
            return App.Placement()

        curve = edge.Curve

        # First we find the translation
        if vtx_type == "Edge" or ignoreVertex:
            # In this case the wanted vertex is the center.
            if curve.TypeId == "Part::GeomCircle":
                center_point = curve.Location
                plc.Base = (center_point.x, center_point.y, center_point.z)
            elif curve.TypeId == "Part::GeomLine":
                edge_points = getPointsFromVertexes(edge.Vertexes)
                line_middle = (edge_points[0] + edge_points[1]) * 0.5
                plc.Base = line_middle
        else:
            vertex = get_element(obj.Shape, vtx)
            if vertex is None:
                return App.Placement()

            plc.Base = (vertex.X, vertex.Y, vertex.Z)

        # Then we find the Rotation
        if curve.TypeId == "Part::GeomCircle":
            plc.Rotation = App.Rotation(curve.Rotation)

        if curve.TypeId == "Part::GeomLine":
            isLine = True
            plane_normal = round_vector(curve.Direction)
            plane_origin = App.Vector(0, 0, 0)
            plane = Part.Plane(plane_origin, plane_normal)
            plc.Rotation = App.Rotation(plane.Rotation)
    elif elt_type == "Face":
        face = get_element(obj.Shape, elt)
        if face is None:
            return App.Placement()

        surface = face.Surface

        # First we find the translation
        if vtx_type == "Face" or ignoreVertex:
            if surface.TypeId == "Part::GeomCylinder":
                centerOfG = face.CenterOfGravity - surface.Center
                centerPoint = surface.Center + centerOfG
                centerPoint = centerPoint + App.Vector().projectToLine(centerOfG, surface.Axis)
                plc.Base = centerPoint
            elif surface.TypeId == "Part::GeomTorus" or surface.TypeId == "Part::GeomSphere":
                plc.Base = surface.Center
            elif surface.TypeId == "Part::GeomCone":
                plc.Base = surface.Apex
            else:
                plc.Base = face.CenterOfGravity
        elif vtx_type == "Edge":
            # In this case the edge is a circle/arc and the wanted vertex is its center.
            edge = get_element(face, vtx)
            if edge is None:
                return App.Placement()

            curve = edge.Curve
            if curve.TypeId == "Part::GeomCircle":
                center_point = curve.Location
                plc.Base = (center_point.x, center_point.y, center_point.z)

            elif (
                surface.TypeId == "Part::GeomCylinder" and curve.TypeId == "Part::GeomBSplineCurve"
            ):
                # handle special case of 2 cylinder intersecting.
                plc.Base = findCylindersIntersection(obj, surface, edge, elt_index)

        else:
            vertex = get_element(obj.Shape, vtx)
            if vertex is None:
                return App.Placement()

            plc.Base = (vertex.X, vertex.Y, vertex.Z)

        # Then we find the Rotation
        if hasattr(surface, "Rotation") and surface.Rotation is not None:
            plc.Rotation = App.Rotation(surface.Rotation)

    # Now plc is the placement relative to the origin determined by the object placement.
    # But it does not take into account Part placements. So if the solid is in a part and
    # if the part has a placement then plc is wrong.

    # change plc to be relative to the object placement.
    plc = obj.Placement.inverse() * plc

    # post-process of plc for some special cases
    if elt_type == "Vertex":
        plc.Rotation = App.Rotation()
    elif isLine:
        plane_normal = round_vector(plc.Rotation.multVec(App.Vector(0, 0, 1)))
        plane_origin = App.Vector(0, 0, 0)
        plane = Part.Plane(plane_origin, plane_normal)
        plc.Rotation = App.Rotation(plane.Rotation)

    # change plc to be relative to the origin of the document.
    # global_plc = getGlobalPlacement(obj, part)
    # plc = global_plc * plc

    # change plc to be relative to the assembly.
    # plc = activeAssembly().Placement.inverse() * plc

    return plc


def get_element(shape, name):
    element = shape.getElement(name)
    if element is None:
        App.Console.PrintWarning(f"Unable to find element {name}.")
    return element


def isRefValid(ref, number_sub):
    if ref is None:
        return False
    if len(ref) != 2:
        return False
    if len(ref[1]) < number_sub:
        return False
    if ref[1][0].find("?") != -1:
        return False

    return True


def round_vector(v, decimals=10):
    """Round each component of the vector to a specified number of decimal places."""
    return App.Vector(round(v.x, decimals), round(v.y, decimals), round(v.z, decimals))


def saveAssemblyPartsPlacements(assembly):
    initialPlcs = {}
    assemblyParts = getMovablePartsWithin(assembly)
    for part in assemblyParts:
        initialPlcs[part.Name] = part.Placement

    return initialPlcs


def restoreAssemblyPartsPlacements(assembly, initialPlcs):
    assemblyParts = getMovablePartsWithin(assembly)
    for part in assemblyParts:
        if part.Name in initialPlcs:
            part.Placement = initialPlcs[part.Name]


def getComAndSize(assembly):
    if assembly.ViewObject is None:
        # these vars use the bounding box which is only available in gui...
        # We could use the real center of mass, but it's too slow to compute it
        return App.Vector(), 100

    bbox = assembly.ViewObject.getBoundingBox()
    if not bbox.isValid():
        return App.Vector(), 100

    com = bbox.Center
    size = bbox.DiagonalLength
    return com, size


def getAssemblyShapes(assembly):
    shapes = []
    assemblyParts = getMovablePartsWithin(assembly)
    for part in assemblyParts:
        shapes.append(part.Shape)

    return shapes


def getJointDistance(joint):
    plc1 = getJcsGlobalPlc(joint.Placement1, joint.Reference1)
    plc2 = getJcsGlobalPlc(joint.Placement2, joint.Reference2)

    # Find the sign
    sign = 1
    plc3 = plc1.inverse() * plc2  # plc3 is plc2 relative to plc1
    if plc3.Base.z < 0:
        sign = -1

    return sign * (plc1.Base - plc2.Base).Length


def getJointXYAngle(joint):
    plc1 = getJcsGlobalPlc(joint.Placement1, joint.Reference1)
    plc2 = getJcsGlobalPlc(joint.Placement2, joint.Reference2)

    plc3 = plc1.inverse() * plc2  # plc3 is plc2 relative to plc1
    x_axis = plc3.Rotation.multVec(App.Vector(1, 0, 0))

    return math.atan2(x_axis.y, x_axis.x)


def getMovingPart(ref):
    if ref is None or len(ref) != 2:
        return None

    obj = ref[0]

    return obj


def getComponentReference(assembly, root_obj, sub_string):
    """
    Takes a full selection path (root + subnames) and normalizes it
    to be rooted at the Assembly Component (Moving Part).

    Returns: (ComponentObject, RelativeSubString) or (None, "")
    """
    if not assembly or not root_obj:
        return None, ""

    doc = assembly.Document

    # 1. Reconstruct full path
    # e.g. ['Part', 'Assembly', 'Cylinder', 'Face1']
    names = [root_obj.Name] + sub_string.split(".")

    # 2. Find Assembly in path
    try:
        asm_idx = names.index(assembly.Name)
    except ValueError:
        return None, ""

    # 3. Identify Component (first valid object after Assembly)
    candidates = names[asm_idx + 1 :]
    if not candidates:
        return None, ""

    component = None
    comp_idx = -1

    for i, obj_name in enumerate(candidates):
        obj = doc.getObject(obj_name)
        if not obj:
            continue

        # Skips (Groups / Flexible Links / LinkGroups / non-geofeature objects)
        if obj.isDerivedFrom("App::DocumentObjectGroup"):
            continue
        if obj.isDerivedFrom("Assembly::AssemblyLink"):
            if hasattr(obj, "Rigid") and not obj.Rigid:
                continue
        if isLinkGroup(obj):
            continue

        if isLink(obj):
            linkedObj = obj.getLinkedObject()
            if linkedObj and not linkedObj.isDerivedFrom("App::GeoFeature"):
                continue
        elif not obj.isDerivedFrom("App::GeoFeature"):
            continue

        component = obj
        comp_idx = asm_idx + 1 + i
        break

    if not component:
        return None, ""

    # 4. Construct new sub-string
    # Everything after the component in the original names list
    relative_parts = names[comp_idx + 1 :]
    new_sub = ".".join(relative_parts)

    return component, new_sub


def truncateSubAtFirst(sub, target):
    # target=part1 & sub=asm.part1.link1.part1.obj -> asm.part1.
    names = sub.split(".")
    sub = ""
    for name in names:
        sub = sub + name + "."
        if name == target:
            break

    return sub


def truncateSubAtLast(sub, target):
    # target=part1 & sub=asm.part1.link1.part1.obj -> asm.part1.link1.part1.
    names = sub.split(".")
    sub = ""
    target_indices = [i for i, name in enumerate(names) if name == target]

    if target_indices:
        last_index = target_indices[-1]
        for i, name in enumerate(names):
            sub += name + "."
            if i == last_index:
                break

    return sub


def swapElNameInSubname(sub_name, new_elName):
    # turns assembly.box.edge1 into assembly.box.new_elName
    names = sub_name.split(".")

    # Replace the last element
    names[-1] = new_elName

    # Join the names back together
    modified_sub = ".".join(names)

    return modified_sub


def addVertexToReference(ref, vertex_name):
    # Turns [obj, ['box.face1']] and 'vertex1' into [obj, ['box.face1', 'box.vertex1']]
    if len(ref) == 2:
        subs = ref[1]
        if len(subs) > 0:
            sub_name = subs[0]
            vertex_full_sub = swapElNameInSubname(sub_name, vertex_name)
            if len(subs) == 2:  # Update the vertex sub
                subs[1] = vertex_full_sub
            else:
                subs.append(vertex_full_sub)

            ref = [ref[0], subs]

    return ref


def createPart(partName, doc):
    if not doc:
        raise ValueError("No active document to add a part to.")

    part = doc.addObject("App::Part", partName)
    body = part.newObject("PartDesign::Body", "Body")
    # Gui.ActiveDocument.ActiveView.setActiveObject('pdbody', body)
    sketch = body.newObject("Sketcher::SketchObject", "Sketch")
    sketch.MapMode = "FlatFace"
    sketch.AttachmentSupport = [(body.Origin.OriginFeatures[3], "")]  # XY_Plane

    # add a circle as a base shape for visualisation
    sketch.addGeometry(Part.Circle(App.Vector(0, 0), App.Vector(0, 0, 1), 5), False)
    doc.recompute()

    return part, body


def getLinkGroup(linkElement):
    if linkElement.TypeId == "App::LinkElement":
        for obj in linkElement.InList:
            if obj.TypeId == "App::Link":
                if linkElement in obj.ElementList:
                    return obj
        print("Link Group not found.")

    return None


def getParentPlacementIfNeeded(part):
    if part.TypeId == "App::LinkElement":
        linkGroup = getLinkGroup(part)
        if linkGroup:
            return linkGroup.Placement

    return Base.Placement()


def generatePropertySettings(documentObject):
    commands = []
    if hasattr(documentObject, "Name"):
        commands.append(f'obj = App.ActiveDocument.getObject("{documentObject.Name}")')
    for propertyName in documentObject.PropertiesList:
        propertyValue = documentObject.getPropertyByName(propertyName)
        propertyType = documentObject.getTypeIdOfProperty(propertyName)
        # Note: OpenCascade precision is 1e-07, angular precision is 1e-05.  For purposes of creating a Macro,
        # we are forcing a reduction in precision so as to get round numbers like 0 instead of tiny near 0 values
        if propertyType == "App::PropertyFloat":
            commands.append(f"obj.{propertyName} = {propertyValue:.5f}")
        elif propertyType == "App::PropertyInt" or propertyType == "App::PropertyBool":
            commands.append(f"obj.{propertyName} = {propertyValue}")
        elif propertyType == "App::PropertyString" or propertyType == "App::PropertyEnumeration":
            commands.append(f'obj.{propertyName} = "{propertyValue}"')
        elif propertyType == "App::PropertyPlacement":
            commands.append(
                f"obj.{propertyName} = App.Placement("
                f"App.Vector({propertyValue.Base.x:.5f},{propertyValue.Base.y:.5f},{propertyValue.Base.z:.5f}),"
                f"App.Rotation(*{[round(n,5) for n in propertyValue.Rotation.getYawPitchRoll()]}))"
            )
        elif propertyType == "App::PropertyXLinkSubHidden":
            commands.append(
                f'obj.{propertyName} = [App.ActiveDocument.getObject("{propertyValue[0].Name}"), {propertyValue[1]}]'
            )
        else:
            # print("Not processing properties of type ", propertyType)
            pass
    return "\n".join(commands) + "\n"



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Class and methods to migrate old FEM App objects

see module end as well as forum topic
https://forum.freecad.org/viewtopic.php?&t=46218
"""

__title__ = "FEM class and methods that migrates old App objects"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

from importlib.util import spec_from_loader

import FreeCAD


class FemMigrateApp:

    def find_spec(self, fullname, path, target=None):
        if fullname in {
            "femsolver.elmer.equations",
            "femsolver.elmer.equations.fluxsolver",
            "femsolver.elmer.solver",
            "femobjects",
            "femobjects._FemConstraintBodyHeatSource",
            "femobjects._FemConstraintElectrostaticPotential",
            "femobjects._FemConstraintFlowVelocity",
            "femobjects._FemConstraintInitialFlowVelocity",
            "femobjects._FemConstraintSelfWeight",
            "femobjects._FemConstraintTie",
            "femobjects._FemElementFluid1D",
            "femobjects._FemElementGeometry1D",
            "femobjects._FemElementGeometry2D",
            "femobjects._FemElementRotation1D",
            "femobjects._FemMaterial",
            "femobjects._FemMaterialMechanicalNonlinear",
            "femobjects._FemMaterialReinforced",
            "femobjects._FemMeshBoundaryLayer",
            "femobjects._FemMeshGmsh",
            "femobjects._FemMeshGroup",
            "femobjects._FemMeshRegion",
            "femobjects._FemMeshResult",
            "femobjects._FemResultMechanical",
            "femobjects._FemSolverCalculix",
            "PyObjects",
            "PyObjects._FemConstraintBodyHeatSource",
            "PyObjects._FemConstraintElectrostaticPotential",
            "PyObjects._FemConstraintFlowVelocity",
            "PyObjects._FemConstraintInitialFlowVelocity",
            "PyObjects._FemConstraintSelfWeight",
            "PyObjects._FemElementFluid1D",
            "PyObjects._FemElementGeometry1D",
            "PyObjects._FemElementGeometry2D",
            "PyObjects._FemElementRotation1D",
            "PyObjects._FemMaterial",
            "PyObjects._FemMaterialMechanicalNonlinear",
            "PyObjects._FemMeshBoundaryLayer",
            "PyObjects._FemMeshGmsh",
            "PyObjects._FemMeshGroup",
            "PyObjects._FemMeshRegion",
            "PyObjects._FemMeshResult",
            "PyObjects._FemResultMechanical",
            "PyObjects._FemSolverCalculix",
            "PyObjects._FemSolverZ88",
            "PyObjects._FemBeamSection",
            "PyObjects._FemFluidSection",
            "PyObjects._FemShellThickness",
            "_FemBeamSection",
            "_FemConstraintSelfWeight",
            "_FemMaterial",
            "_FemMaterialMechanicalNonlinear",
            "_FemMeshGmsh",
            "_FemMeshGroup",
            "_FemMeshRegion",
            "_FemResultMechanical",
            "_FemShellThickness",
            "_FemSolverCalculix",
            "_FemSolverZ88",
            "_FemMechanicalResult",
            "FemResult",
            "_MechanicalMaterial",
            "FemBeamSection",
            "FemShellThickness",
            "MechanicalAnalysis",
            "MechanicalMaterial",
        }:
            return spec_from_loader(fullname, self)
        return None

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        return self.load_module(module)

    def load_module(self, module):
        if module.__name__ == "femsolver.elmer.equations":
            return self
        if module.__name__ == "femsolver.elmer.equations.fluxsolver":
            import femsolver.elmer.equations.flux

            module.Proxy = femsolver.elmer.equations.flux.Proxy
            if FreeCAD.GuiUp:
                module.ViewProxy = femsolver.elmer.equations.flux.ViewProxy

        if module.__name__ == "femsolver.elmer.solver":
            from femobjects.solver_elmer import SolverElmer

            module.Proxy = SolverElmer
            if FreeCAD.GuiUp:
                from femviewprovider.view_solver_elmer import VPSolverElmer

                module.ViewProxy = VPSolverElmer

        if module.__name__ == "femobjects":
            module.__path__ = "femobjects"
        if module.__name__ == "femobjects._FemConstraintBodyHeatSource":
            import femobjects.constraint_bodyheatsource

            module.Proxy = femobjects.constraint_bodyheatsource.ConstraintBodyHeatSource
        if module.__name__ == "femobjects._FemConstraintElectrostaticPotential":
            import femobjects.constraint_electrostaticpotential

            module.Proxy = (
                femobjects.constraint_electrostaticpotential.ConstraintElectrostaticPotential
            )
        if module.__name__ == "femobjects._FemConstraintFlowVelocity":
            import femobjects.constraint_flowvelocity

            module.Proxy = femobjects.constraint_flowvelocity.ConstraintFlowVelocity
        if module.__name__ == "femobjects._FemConstraintInitialFlowVelocity":
            import femobjects.constraint_initialflowvelocity

            module.Proxy = femobjects.constraint_initialflowvelocity.ConstraintInitialFlowVelocity
        if module.__name__ == "femobjects._FemConstraintSelfWeight":
            import femobjects.constraint_selfweight

            module._FemConstraintSelfWeight = femobjects.constraint_selfweight.ConstraintSelfWeight
        if module.__name__ == "femobjects._FemConstraintTie":
            import femobjects.constraint_tie

            module._FemConstraintTie = femobjects.constraint_tie.ConstraintTie
        if module.__name__ == "femobjects._FemElementFluid1D":
            import femobjects.element_fluid1D

            module._FemElementFluid1D = femobjects.element_fluid1D.ElementFluid1D
        if module.__name__ == "femobjects._FemElementGeometry1D":
            import femobjects.element_geometry1D

            module._FemElementGeometry1D = femobjects.element_geometry1D.ElementGeometry1D
        if module.__name__ == "femobjects._FemElementGeometry2D":
            import femobjects.element_geometry2D

            module._FemElementGeometry2D = femobjects.element_geometry2D.ElementGeometry2D
        if module.__name__ == "femobjects._FemElementRotation1D":
            import femobjects.element_rotation1D

            module._FemElementRotation1D = femobjects.element_rotation1D.ElementRotation1D
        if module.__name__ == "femobjects._FemMaterial":
            import femobjects.material_common

            module._FemMaterial = femobjects.material_common.MaterialCommon
        if module.__name__ == "femobjects._FemMaterialMechanicalNonlinear":
            import femobjects.material_mechanicalnonlinear

            module._FemMaterialMechanicalNonlinear = (
                femobjects.material_mechanicalnonlinear.MaterialMechanicalNonlinear
            )
        if module.__name__ == "femobjects._FemMaterialReinforced":
            import femobjects.material_reinforced

            module._FemMaterialReinforced = femobjects.material_reinforced.MaterialReinforced
        if module.__name__ == "femobjects._FemMeshBoundaryLayer":
            import femobjects.mesh_boundarylayer

            module._FemMeshBoundaryLayer = femobjects.mesh_boundarylayer.MeshBoundaryLayer
        if module.__name__ == "femobjects._FemMeshGmsh":
            import femobjects.mesh_gmsh

            module._FemMeshGmsh = femobjects.mesh_gmsh.MeshGmsh
        if module.__name__ == "femobjects._FemMeshGroup":
            import femobjects.mesh_group

            module._FemMeshGroup = femobjects.mesh_group.MeshGroup
        if module.__name__ == "femobjects._FemMeshRegion":
            import femobjects.mesh_region

            module._FemMeshRegion = femobjects.mesh_region.MeshRegion
        if module.__name__ == "femobjects._FemMeshResult":
            import femobjects.mesh_result

            module._FemMeshResult = femobjects.mesh_result.MeshResult
        if module.__name__ == "femobjects._FemResultMechanical":
            import femobjects.result_mechanical

            module._FemResultMechanical = femobjects.result_mechanical.ResultMechanical
        if module.__name__ == "femobjects._FemSolverCalculix":
            import femobjects.solver_ccxtools

            module._FemSolverCalculix = femobjects.solver_ccxtools.SolverCcxTools

        if module.__name__ == "PyObjects":
            module.__path__ = "PyObjects"
        if module.__name__ == "PyObjects._FemConstraintBodyHeatSource":
            import femobjects.constraint_bodyheatsource

            module.Proxy = femobjects.constraint_bodyheatsource.ConstraintBodyHeatSource
        if module.__name__ == "PyObjects._FemConstraintElectrostaticPotential":
            import femobjects.constraint_electrostaticpotential

            module.Proxy = (
                femobjects.constraint_electrostaticpotential.ConstraintElectrostaticPotential
            )
        if module.__name__ == "PyObjects._FemConstraintFlowVelocity":
            import femobjects.constraint_flowvelocity

            module.Proxy = femobjects.constraint_flowvelocity.ConstraintFlowVelocity
        if module.__name__ == "PyObjects._FemConstraintInitialFlowVelocity":
            import femobjects.constraint_initialflowvelocity

            module.Proxy = femobjects.constraint_initialflowvelocity.ConstraintInitialFlowVelocity
        if module.__name__ == "PyObjects._FemConstraintSelfWeight":
            import femobjects.constraint_selfweight

            module._FemConstraintSelfWeight = femobjects.constraint_selfweight.ConstraintSelfWeight
        if module.__name__ == "PyObjects._FemElementFluid1D":
            import femobjects.element_fluid1D

            module._FemElementFluid1D = femobjects.element_fluid1D.ElementFluid1D
        if module.__name__ == "PyObjects._FemElementGeometry1D":
            import femobjects.element_geometry1D

            module._FemElementGeometry1D = femobjects.element_geometry1D.ElementGeometry1D
        if module.__name__ == "PyObjects._FemElementGeometry2D":
            import femobjects.element_geometry2D

            module._FemElementGeometry2D = femobjects.element_geometry2D.ElementGeometry2D
        if module.__name__ == "PyObjects._FemElementRotation1D":
            import femobjects.element_rotation1D

            module._FemElementRotation1D = femobjects.element_rotation1D.ElementRotation1D
        if module.__name__ == "PyObjects._FemMaterial":
            import femobjects.material_common

            module._FemMaterial = femobjects.material_common.MaterialCommon
        if module.__name__ == "PyObjects._FemMaterialMechanicalNonlinear":
            import femobjects.material_mechanicalnonlinear

            module._FemMaterialMechanicalNonlinear = (
                femobjects.material_mechanicalnonlinear.MaterialMechanicalNonlinear
            )
        if module.__name__ == "PyObjects._FemMeshBoundaryLayer":
            import femobjects.mesh_boundarylayer

            module._FemMeshBoundaryLayer = femobjects.mesh_boundarylayer.MeshBoundaryLayer
        if module.__name__ == "PyObjects._FemMeshGmsh":
            import femobjects.mesh_gmsh

            module._FemMeshGmsh = femobjects.mesh_gmsh.MeshGmsh
        if module.__name__ == "PyObjects._FemMeshGroup":
            import femobjects.mesh_group

            module._FemMeshGroup = femobjects.mesh_group.MeshGroup
        if module.__name__ == "PyObjects._FemMeshRegion":
            import femobjects.mesh_region

            module._FemMeshRegion = femobjects.mesh_region.MeshRegion
        if module.__name__ == "PyObjects._FemMeshResult":
            import femobjects.mesh_result

            module._FemMeshResult = femobjects.mesh_result.MeshResult
        if module.__name__ == "PyObjects._FemResultMechanical":
            import femobjects.result_mechanical

            module._FemResultMechanical = femobjects.result_mechanical.ResultMechanical
        if module.__name__ == "PyObjects._FemSolverCalculix":
            import femobjects.solver_ccxtools

            module._FemSolverCalculix = femobjects.solver_ccxtools.SolverCcxTools
        if module.__name__ == "PyObjects._FemSolverZ88":
            import femsolver.z88.solver

            module._FemSolverZ88 = femsolver.z88.solver.Proxy

        if module.__name__ == "PyObjects._FemBeamSection":
            import femobjects.element_geometry1D

            module._FemBeamSection = femobjects.element_geometry1D.ElementGeometry1D
        if module.__name__ == "PyObjects._FemFluidSection":
            import femobjects.element_fluid1D

            module._FemFluidSection = femobjects.element_fluid1D.ElementFluid1D
        if module.__name__ == "PyObjects._FemShellThickness":
            import femobjects.element_geometry2D

            module._FemShellThickness = femobjects.element_geometry2D.ElementGeometry2D

        if module.__name__ == "_FemBeamSection":
            import femobjects.element_geometry1D

            module._FemBeamSection = femobjects.element_geometry1D.ElementGeometry1D
        if module.__name__ == "_FemConstraintSelfWeight":
            import femobjects.constraint_selfweight

            module._FemConstraintSelfWeight = femobjects.constraint_selfweight.ConstraintSelfWeight
        if module.__name__ == "_FemMaterial":
            import femobjects.material_common

            module._FemMaterial = femobjects.material_common.MaterialCommon
        if module.__name__ == "_FemMaterialMechanicalNonlinear":
            import femobjects.material_mechanicalnonlinear

            module._FemMaterialMechanicalNonlinear = (
                femobjects.material_mechanicalnonlinear.MaterialMechanicalNonlinear
            )
        if module.__name__ == "_FemMeshGmsh":
            import femobjects.mesh_gmsh

            module._FemMeshGmsh = femobjects.mesh_gmsh.MeshGmsh
        if module.__name__ == "_FemMeshGroup":
            import femobjects.mesh_group

            module._FemMeshGroup = femobjects.mesh_group.MeshGroup
        if module.__name__ == "_FemMeshRegion":
            import femobjects.mesh_region

            module._FemMeshRegion = femobjects.mesh_region.MeshRegion
        if module.__name__ == "_FemResultMechanical":
            import femobjects.result_mechanical

            module._FemResultMechanical = femobjects.result_mechanical.ResultMechanical
        if module.__name__ == "_FemShellThickness":
            import femobjects.element_geometry2D

            module._FemShellThickness = femobjects.element_geometry2D.ElementGeometry2D
        if module.__name__ == "_FemSolverCalculix":
            import femobjects.solver_ccxtools

            module._FemSolverCalculix = femobjects.solver_ccxtools.SolverCcxTools
        if module.__name__ == "_FemSolverZ88":
            import femsolver.z88.solver

            module._FemSolverZ88 = femsolver.z88.solver.Proxy

        if module.__name__ == "_FemMechanicalResult":
            import femobjects.result_mechanical

            module._FemMechanicalResult = femobjects.result_mechanical.ResultMechanical
        if module.__name__ == "FemResult":
            import femobjects.result_mechanical

            module.FemResult = femobjects.result_mechanical.ResultMechanical
        if module.__name__ == "_MechanicalMaterial":
            import femobjects.material_common

            module._MechanicalMaterial = femobjects.material_common.MaterialCommon

        if module.__name__ == "FemBeamSection":
            import femobjects.element_geometry1D

            module._FemBeamSection = femobjects.element_geometry1D.ElementGeometry1D
            if FreeCAD.GuiUp:
                import femviewprovider.view_element_geometry1D

                module._ViewProviderFemBeamSection = (
                    femviewprovider.view_element_geometry1D.VPElementGeometry1D
                )
        if module.__name__ == "FemShellThickness":
            import femobjects.element_geometry2D

            module._FemShellThickness = femobjects.element_geometry2D.ElementGeometry2D
            if FreeCAD.GuiUp:
                import femviewprovider.view_element_geometry2D

                module._ViewProviderFemShellThickness = (
                    femviewprovider.view_element_geometry2D.VPElementGeometry2D
                )
        if module.__name__ == "MechanicalAnalysis":
            import femobjects.base_fempythonobject

            module._FemAnalysis = femobjects.base_fempythonobject.BaseFemPythonObject
            if FreeCAD.GuiUp:
                import femviewprovider.view_base_femobject

                module._ViewProviderFemAnalysis = (
                    femviewprovider.view_base_femobject.VPBaseFemObject
                )
        if module.__name__ == "MechanicalMaterial":
            import femobjects.material_common

            module._MechanicalMaterial = femobjects.material_common.MaterialCommon
            if FreeCAD.GuiUp:
                import femviewprovider.view_material_common

                module._ViewProviderMechanicalMaterial = (
                    femviewprovider.view_material_common.VPMaterialCommon
                )
        return None


"""
possible entries in the old files:
(the class name in the old file does not matter, we ever only had one class per module)

further renaming objects
module="femsolver.elmer.equations.fluxsolver"
in this modules object class and viewprovider class are in same module


fourth big moving
renaming class and module names in femobjects
TODO add link to commit before the first commit
module="femobjects._FemConstraintBodyHeatSource"
module="femobjects._FemConstraintElectrostaticPotential"
module="femobjects._FemConstraintFlowVelocity"
module="femobjects._FemConstraintInitialFlowVelocity"
module="femobjects._FemConstraintSelfWeight"
module="femobjects._FemConstraintTie"
module="femobjects._FemElementFluid1D"
module="femobjects._FemElementGeometry1D"
module="femobjects._FemElementGeometry2D"
module="femobjects._FemElementRotation1D"
module="femobjects._FemMaterial"
module="femobjects._FemMaterialMechanicalNonlinear"
module="femobjects._FemMaterialReinforced"
module="femobjects._FemMeshBoundaryLayer"
module="femobjects._FemMeshGmsh"
module="femobjects._FemMeshGroup"
module="femobjects._FemMeshRegion"
module="femobjects._FemMeshResult"
module="femobjects._FemResultMechanical"
module="femobjects._FemSolverCalculix"

third big moving
from PyObjects to femobjects, following the parent commit
https://github.com/FreeCAD/FreeCAD/tree/07ae0e56c4/src/Mod/Fem/PyObjects
module="PyObjects._FemConstraintBodyHeatSource"
module="PyObjects._FemConstraintElectrostaticPotential"
module="PyObjects._FemConstraintFlowVelocity"
module="PyObjects._FemConstraintInitialFlowVelocity"
module="PyObjects._FemConstraintSelfWeight"
module="PyObjects._FemElementFluid1D"
module="PyObjects._FemElementGeometry1D"
module="PyObjects._FemElementGeometry2D"
module="PyObjects._FemElementRotation1D"
module="PyObjects._FemMaterial"
module="PyObjects._FemMaterialMechanicalNonlinear"
module="PyObjects._FemMeshBoundaryLayer"
module="PyObjects._FemMeshGmsh"
module="PyObjects._FemMeshGroup"
module="PyObjects._FemMeshRegion"
module="PyObjects._FemMeshResult"
module="PyObjects._FemResultMechanical"
module="PyObjects._FemSolverCalculix"
module="PyObjects._FemSolverZ88"

renamed between the second and third big moveings
module="PyObjects._FemBeamSection"
module="PyObjects._FemFluidSection"
module="PyObjects._FemShellThickness"

second big moveing
into PyObjects, following the parent commit
https://github.com/FreeCAD/FreeCAD/tree/7f884e8bff/src/Mod/Fem
module="_FemBeamSection"
module="_FemConstraintSelfWeight"
module="_FemMaterial"
module="_FemMaterialMechanicalNonlinear."
module="_FemMeshGmsh"
module="_FemMeshGroup"
module="_FemMeshRegion"
module="_FemResultMechanical"
module="_FemShellThickness"
module="_FemSolverCalculix"
module="_FemSolverZ88"

renamed between the first and second big moveings
module="_FemMechanicalResult"
module="FemResult"
module="_MechanicalMaterial"

first big moving
split modules from one module into make, obj class, vp class, command
new obj class module names had a _
following the parent commit of the first split commit
https://github.com/FreeCAD/FreeCAD/tree/c3328d6b4e/src/Mod/Fem
in this modules there where object class and viewprovider class together
module="FemBeamSection"
module="FemShellThickness"
module="MechanicalAnalysis"
module="MechanicalMaterial"
"""



# ***************************************************************************
# *   Copyright (c) 2024 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD
import FreeCADGui

""" Test active object list """


class TestActiveObject(unittest.TestCase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("PartDesignTestSketch")
        self.doc.UndoMode = True

    def testPartBody(self):
        self.doc.openTransaction("Create part")
        part = self.doc.addObject("App::Part", "Part")
        FreeCADGui.activateView("Gui::View3DInventor", True)
        FreeCADGui.activeView().setActiveObject("part", part)
        self.doc.commitTransaction()

        self.doc.openTransaction("Create body")
        body = self.doc.addObject("PartDesign::Body", "Body")
        part.addObject(body)
        FreeCADGui.activateView("Gui::View3DInventor", True)
        FreeCADGui.activeView().setActiveObject("pdbody", body)
        self.doc.commitTransaction()

        self.doc.undo()  # undo body creation
        self.doc.undo()  # undo part creation

        FreeCADGui.updateGui()

        self.doc.openTransaction("Create body")
        body = self.doc.addObject("PartDesign::Body", "Body")
        FreeCADGui.activateView("Gui::View3DInventor", True)
        FreeCADGui.activeView().setActiveObject("pdbody", body)
        self.doc.commitTransaction()

        FreeCADGui.updateGui()

    def tearDown(self):
        FreeCAD.closeDocument("PartDesignTestSketch")
